"""
Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""
from enum import Enum

radio_mode_str = '0: 1Mbit/s Nordic, 1: 2Mbit/s Nordic, 3: 1Mbit/s BLE, 3: 2Mbit/s BLE'


class CommandDesc(Enum):

    # information commands:
    version = {'args': None, 'desc': 'returns the Gateway Firmware version'}
    print_config = {'args': None, 'desc': 'returns the current Gateway configuration - short list'}
    print_config_extended =  {'args': None, 'desc': 'returns the current Gateway configuration - detailed list'}
    get_device_address = {'args': None, 'desc': 'returns the Gateway device address, a unique address per device'}
    get_name = {'args': None, 'desc': 'returns the Gateway name, a unique name per device'}
    full_packet_mode = {'args': ['IS_FULL_PACKET_MODE'], 'desc': 'relevant only for LEGACY (BLE4) packets. if enabled the Gateway passes the BLE4 packet inclusing the physical header, default 0 (disable), e.g. 1'}
    run_ed_sample = {'args': None, 'desc': 'returns the Energy power the Gateway detect over the air using nordic ED methods'}

    # configurations commands:
    set_energizing_pattern = {'args': ['ENERGY_PATTERN'], 'desc': 'set the Gateway energy pattern, default 18'}
    time_profile = {'args': ['CYCLE_TIME_MS', 'ON_TIME_MS'], 'desc': 'set the Gateway time profile - cycle + on/transmit time, default 15 5'}
    scan_ch = {'args': ['RX_CHANNEL_FREQ', 'WHITENING_CHANNEL'], 'desc': 'set the Gateway scan/recieve channel or frequency, default 37, e.g. 10, 2480, 39, ... the WHITENING_CHANNEL is optional if the data whitening channel is different than the scan channel.'}
    set_scan_radio = {'args': ['SCAN_RADIO_MODE', 'PREAMBLE_LENGTH'], 'desc': f'set the Gateway scan radio mode [{radio_mode_str}], default 3, and the preamble length [0=8bits, 1=16bits], default 0'}
    output_power = {'args': ['OUTPUT_POWER_STR'], 'desc': 'set the Gateway intenral output-power, default pos3dBm, e.g. neg4dBm, ...'}
    bypass_pa = {'args': ['IS_BYPASS_PA_TX'], 'desc': 'set the Gateway TX Power Amplifier: 0 = means no pybass, hence PA is on, 1 = otherwise, default 0, e.g. 1'}
    bypass_tx_rx = {'args': ['IS_BYPASS_PA_TX', 'IS_BYPASS_PA_RX'], 'desc': 'set the Gateway TX and RX Power Amplifier: 0 = means no pybass, hence PA is on, 1 = otherwise, default 0, e.g. 1'}
    beacons_backoff = {'args': ['BEACONS_BACKOF_DBM'], 'desc': 'set the Gateway beacons backoff/power reduction from the maximum power in dbm, default 0 = means no beacons backoff, e.g. 2'}

    set_sub_1_ghz_energizing_frequency = {'args': ['SUB1G_FREQ_MHZ'], 'desc': 'set the Gateway sub1g frequency in MHz, default 915000'}
    set_sub_1_ghz_energizing_mode = {'args': ['SUB1G_ENERGY_MODE'], 'desc': 'set the Gateway sub1g energy mode: 0 = [default] energy on single frequency based on the SUB1G_FREQ_MHZ, 1 = enrgy using the FCC hopping method, 905Mhz till 920 Mhz with 300 Khz intervals'}
    set_sub_1_ghz_power = {'args': ['SUB1G_POWER_DBM'], 'desc': 'set the Gateway sub1g output power in dBm, between 17-29, default 29'}
    sub1g_sync = {'args': ['IS_SUB1G_SYNCED'], 'desc': 'set the Gateway sub1g transmission to start at the same time the 2.4 transmission starts, otherwise the sub1g trasmission would start at some point of the 2.4 cycle. default 1, e.g. 0'}
    set_sub_1_ghz_starting_point = {'args': ['SUB1G_START_POINT_MODE'], 'desc': 'set the Gateway sub1g time sync w.r.t the 2.4 transmission. mode 0 = starts with the first beacon [default], 1 = starts after completing beacons in 2.4 transmission. defaul 0, e.g. 1'}

    listen_to_tag_only = {'args': ['DO_LISTEN_TO_TAG_ONLY'], 'desc': 'enable/disable listen to packets that were echoed by bridges. 1 = only from tags [defualt], 0 = also packets echo from bridges'}
    set_rssi_th = {'args': ['RSSI_THRESHOLD'], 'desc': 'set the Gateway rssi threshold to filter packets with low signal. 0 [default] meaning no filtering, otherwise value is in minus, so to filter all packets with rssi(signal power) lower than -50dBm, set the threshold to 50'}

    pl_gw_config = {'args': ['ENABLE_PRODUCTION_LINE'], 'desc': 'enable/disable production line mode - when pulse detected on GPIO P010, Gateway sends message, start with constant wave for PL_DELAY time [defualt is 0] and continues with the configured energy pattern. default 0, e.g. 1'}
    trigger_pl = {'args': None, 'desc': 'trigger the Production Line process similar to extenral triggering by GPIO'}
    set_pl_delay = {'args': ['PL_DELAY_MS'], 'desc': 'set the Gateway Constant Wave duration in ms upon triggering the Production Line process. default 0, e.g. 100'}

    enable_crc = {'args': ['DO_PASS_BAD_CRC_PACKET'], 'desc': 'enable the Gateway to pass bad crc packets, default is 0 (disabled), e.g. 1'}
    print_crc = {'args': None, 'desc': 'returns the crc packets (any packets not only Wiliot) counter'}
    reset_crc = {'args': None, 'desc': 'reset the crc packets counters'}

    set_pacer_interval = {'args': ['PACER_INTERVAL_S'], 'desc': 'set the Gateway pacer interval - the minimum time in seconds the gateway should pass a packet from specific tag, default 0 (meaining no pacing), e.g 1'}
    set_packet_filter_on = {'args': None, 'desc': 'enable packet filter, which means the Gateway passes only unique packets'}
    set_packet_filter_off = {'args': None, 'desc': '[default] disable packet filter, which means the Gateway passes only unique packets'}

    # application commands:
    gateway_app = {'args': None, 'desc': 'start the Gateway application to transmit energy/beacons and receive packets'}
    reset = {'args': None, 'desc': 'soft reset the Gateway - will restore default configuration and re-run all the processes'}
    cancel = {'args': None, 'desc': 'stop the Gateway application, no transmiting nor receiving'}

    # Management commands:
    store_to_flash = {'args': None, 'desc': 'store the current Gateway configuration to its flash, so even after power-down or software reset the Gateway is configured with the saved configurations'}
    move_to_bootloader = {'args': None, 'desc': 'move the Gateway to bootloader mode for firmware upgrade purposes'}
    enable_hw_dual_band = {'args': ['IS_DUAL_BAND'], 'desc': 'enable or disable dual band options (sub1g) for the Gateway. the default is dual band enabled, e.g. 1 or 0'}

    # Utils:
    start_2_4_ghz_energizing = {'args': ['ENRGY_FREQ, RADIO_MODE'], 'desc': f'start to energize with constant wave on the selected freq and radio mode [{radio_mode_str}], e.g. 2480, 0'}
    start_rx = {'args': None, 'desc': f'start to listen only based on the configured scan channel'}
    start_cw = {'args': ['ON_TIME_MS, OFF_TIME_MS'], 'desc': 'start Gateway transmitting (NO recieiving at all) of constant wave for the specified on time in ms, and stop transmitting based on the specified off time in ms, e.g. 2 5'}
    enable_hv_gpio = {'args': None, 'desc': 'enable the high voltage GPIO on the Gateway. when this mode is enable, instead of energy transmitted by the Gateway, the Gateway set GPIO P009 to High to interact with external hardware'}
    set_lbt = {'args': ['DO_LISTEN_BEFORE_TALK'], 'desc': 'enable the option to transmit only if there is no noise (under -70dBm) over the air. default 0, e.g. 1'}

    # dynamic patterns
    beacons_train = {'args': ['IS_SUB1G_ENERGY', 'NUMBER_OF_BEACONS', 'BEACON1', 'BEACON2', '...'], 'desc': 'set dynamic energy pattern on the Gateway (the EP number is always 41). the command set if there is energy in sub1g during the on time that is configured based on the time profile (different command), each beacon setup contains the frequncy,beacon duration,silence time till the next beacon. e.g. EP18 is the same as: !beacons_train 0 4 2402,550,1 2426,550,1 2480,550,1 2480,3246,1'}

    # gpio
    cmd_gpio = {'args': ['COMMAND', 'PIN', 'ARGUMENT'], 'desc': 'for detailed explanation please check https://wiliot.atlassian.net/wiki/spaces/SW/pages/3228827713/GW+Tester+Board+GPIO. examples: !cmd_gpio SEND 1 P004 pulse 1 500, !cmd_gpio SEND 1 P004 static 1, !cmd_gpio CONTROL_OUT P030 0, !cmd_gpio CONTROL_IN P030 1'}



"""
Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""


import shutil
from os import popen
from os.path import join, isdir
from sys import platform
from enum import Enum
import os
import json
try:
    from tkinter import messagebox
except ModuleNotFoundError as e:
    print(f'could not import tkinter: {e}')


class WiliotPackages(Enum):
    CLOUD = ['pywiliot-api']
    CORE = ['pywiliot-core', 'pywiliot-api']
    TOOLS = ['pywiliot-tools', 'pywiliot-core', 'pywiliot-api']
    DEPLOYMENT = ['pywiliot-deployment-tools', 'pywiliot-tools', 'pywiliot-core', 'pywiliot-api']
    TESTERS = ['pywiliot-testers', 'pywiliot-tools', 'pywiliot-core', 'pywiliot-api']


def update_internal_wiliot_packages(wiliot_repo=WiliotPackages.CLOUD, branch_name='master',
                                    overwrite=False, no_wiliot_dep=False,
                                    username_bitbucket=None, password_bitbucket=None,
                                    install_from_local_folder='', install_from_git_folder='',
                                    branch_name_for_dep='master'):
    # check platform command prefix
    python_ver = ''
    if platform == "darwin" or platform == 'linux':
        # macos or linux
        python_ver = '3'
    # install pywiliot and requirements:
    # update dependencies
    print("----------------------------------\nupdate pip:\n----------------------------------")
    p = popen(f'python{python_ver} -m pip install --upgrade pip')
    rsp = p.read()
    print(rsp)
    pkg_versions = []
    req_status = []
    for repo_num, repo_name in enumerate(wiliot_repo.value):
        need_to_install = True
        package_name = repo_name.replace('py', '')
        folder_name = package_name.replace('-', '_')
        print('check if package already exist')
        p = popen(f'pip{python_ver} show {package_name}')
        rsp = p.read()
        print(rsp)
        if package_name in rsp:
            if overwrite:
                print(f'--------------------------------------------------------------------\n'
                      f'uninstall existing package: {package_name}\n'
                      f'--------------------------------------------------------------------\n')
                p = popen(f'pip{python_ver} uninstall {package_name} -y')
            else:
                print(f'--------------------------------------------------------------------\n'
                      f'for uninstall existing package: {package_name}, please press ENTER\n'
                      f'--------------------------------------------------------------------\n')
                p = popen(f'pip{python_ver} uninstall {package_name}')
            rsp = p.read()
            print(rsp)
        print(f'--------------------------------------------------------------------\n'
              f'update wiliot package: {package_name}\n'
              f'--------------------------------------------------------------------')
        if install_from_local_folder or install_from_git_folder:
            local_repo_path = join(install_from_local_folder if install_from_local_folder else install_from_git_folder,
                                   repo_name)
            if isdir(local_repo_path):
                p = popen(f'pip{python_ver} install -e {local_repo_path}  --upgrade')
                rsp = p.read()
                print(rsp)
                need_to_install = False
            else:
                print(f'{repo_name} does not exist in local folder, try to install from bitbucket')

        if need_to_install:
            if username_bitbucket is None and password_bitbucket is None:
                if repo_num == 0:
                    p = popen(f'pip{python_ver} install git+ssh://git@bitbucket.org/wiliot/{repo_name}.git'
                              f'@{branch_name}#egg={folder_name} --upgrade')
                else:
                    p = popen(f'pip{python_ver} install git+ssh://git@bitbucket.org/wiliot/{repo_name}.git'
                              f'@{branch_name_for_dep}#egg={folder_name} --upgrade')
                rsp = p.read()
                print(rsp)
                if 'Successfully installed' not in rsp.split('\n')[-2]:
                    print(
                        f'problem with ssh key credentials. Please rerun the script using username (-u) and password (-p).'
                        f'\nTo extract Attlasian username and password go to '
                        f'https://wiliot.atlassian.net/wiki/spaces/SW/pages/2890662124/'
                        f'PyWiliot+Installation+Guidelines+-+Private#using-username-and-password%3A')
                    return
            elif username_bitbucket is not None and password_bitbucket is not None:
                if repo_num == 0:
                    p = popen(
                        f'pip{python_ver} install git+https://{username_bitbucket}:{password_bitbucket}@bitbucket.org'
                        f'/wiliot/{repo_name}.git@{branch_name}#egg={folder_name} --upgrade')
                else:
                    p = popen(
                        f'pip{python_ver} install git+https://{username_bitbucket}:{password_bitbucket}@bitbucket.org'
                        f'/wiliot/{repo_name}.git@{branch_name_for_dep}#egg={folder_name} --upgrade')
                rsp = p.read()
                print(rsp)
            else:
                print('you Must specified both username and password or do NOT specified neither '
                      'for installation using SSH key')
                return

        print(f'check wiliot package: {package_name}')
        p = popen(f'pip{python_ver} show {package_name}')
        rsp = p.read()
        print(rsp)
        if 'Location' in rsp:
            pkg_versions.append(rsp.split('Version: ')[1].split('\n')[0])
            req_path = rsp.split('Location: ')[-1].split('\n')[0]
            if install_from_git_folder:
                if 'Editable project location: ' in rsp:
                    req_path = rsp.split('Editable project location: ')[-1].split('\n')[0]
            elif install_from_local_folder:
                site_packages_path = join(req_path, folder_name)
                if isdir(site_packages_path):
                    shutil.rmtree(site_packages_path)
                shutil.copytree(join(install_from_local_folder, repo_name, folder_name), site_packages_path)
                req_path = join(install_from_local_folder, repo_name)

            print("--------------------------------------------------------------------\n"
                  "update dependencies:\n"
                  "--------------------------------------------------------------------")
            requirements_path = join(req_path, folder_name, 'requirements.txt')
            p = popen(f'pip{python_ver} install -r "{requirements_path}"')
            rsp = p.read()
            print(rsp)
            req_status.append('ERROR' not in rsp)
        else:
            pkg_versions.append('')
            req_status.append(False)

        if no_wiliot_dep:
            break

    print('done installation')
    print("************************************************************************\n"
          "********************            Summary            *********************\n"
          "************************************************************************")
    for i, repo_name in enumerate(wiliot_repo.value):
        package_name = repo_name.replace('py', '')
        is_installed = f'installed [{pkg_versions[i]}]' if pkg_versions[i] else 'NOT installed'
        sum_str = f'{package_name} was {is_installed}'
        if is_installed:
            req_is_installed = 'installed' if req_status[i] else 'NOT installed'
            sum_str += f'and the dependencies were {req_is_installed}'
        print(sum_str)
        if no_wiliot_dep:
            break


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description='Update Wiliot Internal Packages')
    parser.add_argument('-p', '--package_name', help='The package name. valid option(wiliot-api, wiliot-core, '
                                                     'wiliot-tools, wiliot-testers, wiliot-deployment-tools')
    parser.add_argument('-b', '--branch_name', help='The branch name you want to use. the default is master',
                        default='master')
    parser.add_argument('-w', '--overwrite', help='If specified, all wiliot existed packages will be overwrite '
                                                  'with the new installation', default=False, action='store_true')
    parser.add_argument('-n', '--no_wiliot_dep', help='If specified, install the main package without wiliot '
                                                      'dependencies packages', default=False, action='store_true')
    parser.add_argument('-u', '--username', help='If specified, install based username and password and not SHH key',
                        default=None)
    parser.add_argument('-pass', '--password', help='If specified, install based username and password and not SHH key',
                        default=None)
    parser.add_argument('-l', '--local_folder', help='If specified, install packages from the specified local folder')
    parser.add_argument('-g', '--git_folder', help='FOR DEVELOPERS. If specified, install packages from the specified'
                                                   'folder, which assumed to be your git workspace')
    parser.add_argument('-b-for-dep', '--branch_for_dependencies',
                        help='if specified, all wiliot package dependencies will be installed for this branch '
                             '(default master)', default='master')
    args = parser.parse_args()
    package_name = args.package_name
    branch_name = args.branch_name
    is_overwrite = args.overwrite
    is_no_wiliot_dep = args.no_wiliot_dep
    username = args.username
    password = args.password
    local_folder = args.local_folder
    git_folder = args.git_folder
    branch_for_dep = args.branch_for_dependencies
    err_msg = 'please add package name, e.g. update_wiliot_packages.py -p wiliot-api.\n' \
              'The valid packages names are: wiliot-api, wiliot-core, wiliot-tools, wiliot-testers, ' \
              'wiliot-deployment-tools'
    if package_name != None:
        try:
            repo_name = None
            for package in WiliotPackages.__members__.values():
                if package_name in package.value[0]:
                    repo_name = package
                    break
            if repo_name is None:
                print(err_msg)
            else:
                update_internal_wiliot_packages(wiliot_repo=repo_name, branch_name=branch_name,
                                                overwrite=is_overwrite, no_wiliot_dep=is_no_wiliot_dep,
                                                username_bitbucket=username, password_bitbucket=password,
                                                install_from_local_folder=local_folder,
                                                install_from_git_folder=git_folder,
                                                branch_name_for_dep=branch_for_dep)
        except Exception as e:
            print('Error during update packages  [ERROR: {}]'.format(e))

    else:
        print('package must be specified')



"""
Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import datetime
import logging
import numpy as np
import pandas as pd
import copy
from enum import Enum
from wiliot_core.packet_data.config_files.parse_packet import *
from wiliot_core.utils.utils import valid_packet_start
from wiliot_core.packet_data.config_files.packet_flow_parameters import get_flow_param


FULL_PACKET_PREFIX = 'full_packet'
LEGACY_PACKET_PREFIX = 'process_packet'


packet_tag_length = 74
MAX_STAT_PARAM_VAL = 65535  # 2 bytes
MAX_GW_CLOCK = 32767  # 7 bits
MAX_TIME_DIFF = 30  # sec


class InlayTypes(Enum):
    TEO_086 = '086'
    TIKI_096 = '096'
    TIKI_099 = '099'
    BATTERY_107 = '107'
    TIKI_117 = '117'
    TIKI_118 = '118'
    TIKI_121 = '121'
    TIKI_122 = '122'
    TIKI_136 = '136'
    TIKI_H_160 = '160'
    TIKI_160 = '160'
    TIKI_168 = '168'
    TIKI_169 = '169'


class Packet(object):
    """
    Wiliot Packet Object

        :param raw_packet: the raw packet to create a Packet object
        :type raw_packet: str or dict

        :return:
    """

    def __init__(self, raw_packet, time_from_start=None, custom_data=None,
                 inlay_type=None, logger_name=None, ignore_crc=False):
        """
        :param raw_packet:
        :type raw_packet: str or dict
        :param logger_name: the log file we want to write into it.
        :type time_from_start: str
        :param time_from_start: the time when the packet was received according to the gw timer
        :type time_from_start: float
        :param custom_data: the packet custom data as dictionary
        :type custom_data: dict
        :param inlay_type: the antenna/ inlay type (e.g. tiki, teo)
        :type inlay_type: str
        :type ignore_crc: bool
        :param ignore_crc: if True, update for all packets crc_valid = True,
                           relevant for data collected by old GW FW version
        """

        self.is_valid_packet = False
        if logger_name is not None:
            self.logger = logging.getLogger(logger_name)
        else:
            self.logger = None

        # init packet input:
        packet_in = None
        time_in = None
        self.gw_process = False
        if type(raw_packet) is dict and raw_packet['is_valid_tag_packet']:
            packet_in = raw_packet['packet']
            time_in = raw_packet['time_from_start']
            self.gw_process = True
        elif type(raw_packet) is str:
            packet_in = raw_packet
            time_in = time_from_start

        # check packet type:
        if FULL_PACKET_PREFIX in packet_in:
            self.is_full_packet = True
        elif LEGACY_PACKET_PREFIX in packet_in:
            self.is_full_packet = False
        else:
            self.is_full_packet = None

        # parse packet
        try:
            packet_in_strip = valid_packet_start(packet_in) or packet_in
            self.gw_data, self.packet_data = parse_packet(packet_in_strip, time_in, self.is_full_packet, ignore_crc)

            # assign results
            self.is_valid_packet = True
        except Exception as e:
            self.printing_and_logging(f'Could not parse the following packet {packet_in} due to {e}')
            self.is_valid_packet = False

        # add more info to packet struct
        self.decoded_data = {}
        if custom_data is not None:
            self.custom_data = custom_data
        else:
            self.custom_data = {}
        if 'timestamp' not in self.custom_data.keys():
            self.custom_data['timestamp'] = datetime.datetime.now().timestamp() * 1000  # ms
        self.inlay_type = inlay_type

        if self.is_valid_packet:
            if self.packet_data['ble_type'] != 'LEGACY':
                self.custom_data['adi_per_packet'] = [self.packet_data['adi']]
                self.packet_data['raw_packet'] = construct_packet_from_fields(packet_data=self.packet_data,
                                                                              fields_to_zero=['adi'])

    def __len__(self):
        """
        gets number of sprinkler occurrences in packet
        """
        return self.gw_data['rssi'].size

    def __eq__(self, packet):
        """
        packet comparison
        """
        if self.is_same_sprinkler(packet):
            if len(packet) == self.__len__():
                return all([np.take(packet.gw_data['gw_packet'], i) == np.take(self.gw_data['gw_packet'], i)
                            for i in range(len(packet))])
        return False

    def __str__(self):
        """
        packet print method
        """
        return str(
            'packet_data={packet_data}, gw_data={gw_data}'.format(packet_data=self.packet_data, gw_data=self.gw_data))

    def printing_and_logging(self, message):
        if self.logger:
            self.logger.info(message)
        else:
            print(message)

    def is_packet_from_bridge(self):
        return self.packet_data['is_packet_from_bridge']

    def get_payload(self):
        if self.is_valid_packet:
            return self.packet_data['data_uid'] + self.packet_data['raw_group_id'] + self.packet_data['nonce'] + \
                   self.packet_data['enc_uid'] + self.packet_data['mic'] + self.packet_data['enc_payload']
        else:
            return ''

    def get_group_id(self):
        return self.packet_data['group_id']

    def set_inlay_type(self, inlay_type):
        inlay_type = inlay_type if isinstance(inlay_type, str) else inlay_type.value
        self.inlay_type = inlay_type

    def is_in(self, packet):
        """
        is packet contains another packet

        :param packet: the other packet to verify
        :type packet: Packet

        :return: bool
        """
        if self.is_same_sprinkler(packet):
            if np.take(packet.gw_data['gw_packet'], 0) in self.gw_data['gw_packet']:
                return True
        return False

    def get_packet(self):
        """
        gets raw packet string
        """
        return str(self.packet_data['raw_packet'])

    def split_packet(self, index):
        """
        split packet by index
        """
        packet_a = self.copy()
        packet_b = self.copy()
        remain = len(self) - index
        for key in self.gw_data.keys():
            for i in range(index):
                packet_a.gw_data[key] = np.delete(packet_a.gw_data[key], -1)

            for i in range(remain):
                packet_b.gw_data[key] = np.delete(packet_b.gw_data[key], 0)

        return packet_a, packet_b

    def copy(self):
        return copy.deepcopy(self)

    def sort(self):
        """
        sort gw_data lists according to gw_time
        """
        isort = np.argsort(self.gw_data['gw_clock'])
        for key in self.gw_data.keys():
            self.gw_data[key] = self.gw_data[key][isort]

    def get_average_rssi(self):
        return np.average(self.gw_data['rssi'])

    def get_packet_size(self):
        return int(self.packet_data['packet_length'])

    def is_short_packet(self):
        return len(self.get_packet_string(process_packet=False)) < self.get_packet_size()

    def get_packet_content(self, raw_packet=None, get_gw_data=False, sprinkler_num=0):
        if raw_packet is None:
            parsed_packet = self
        else:
            raw_packet = valid_packet_start(raw_packet) or raw_packet
            parsed_packet = Packet(raw_packet=raw_packet)

        if get_gw_data:
            return parsed_packet.packet_data['raw_packet'] + parsed_packet.gw_data['gw_packet'].take(sprinkler_num)

        return parsed_packet.packet_data['raw_packet']

    def get_packet_string(self, i=0, gw_data=True, process_packet=True, full_packet=False):
        """
        gets process_packet string
        """
        if not self.is_valid_packet:
            return None

        process_packet_string = ['', '']
        if process_packet:
            if full_packet:
                raise ValueError('Could not get packet string since both full and process packet flags are true')
            process_packet_string = ['process_packet("', '")']
        if full_packet:
            process_packet_string = ['full_packet("', '")']

        if gw_data:
            gw_data = self.gw_data['gw_packet'].take(i)
        else:
            gw_data = ''
        return '{raw_packet}{gw_data}'.format(raw_packet=self.packet_data['raw_packet'],
                                              gw_data=gw_data).join(process_packet_string)

    def get_adva(self):
        return self.packet_data['adv_address']

    def get_flow(self):
        flow_version = hex(int(self.packet_data['flow_ver'], 16))
        return flow_version

    def get_rssi(self):
        return str(self.gw_data['rssi'])

    def get_packet_data_names(self):
        """
        extract all keys name which can potentially be a column in packet df
        :return:
        :rtype: list
        """
        dict_temp = self.__dict__
        keys = list(dict_temp.keys())
        for k in dict_temp.keys():
            if isinstance(dict_temp[k], dict):
                keys = keys + list(dict_temp[k].keys())
        return keys

    def is_same_sprinkler(self, packet):
        raw_packet = packet.packet_data['raw_packet']
        if raw_packet == self.packet_data['raw_packet']:
            return True
        return False

    def append_to_sprinkler(self, packet, output_log=True, do_sort=False):
        status = True
        if self.is_same_sprinkler(packet):
            for i in range(len(packet)):
                try:
                    # stat param should be unique (time + rssi for same packet)- we make sure no duplications are added.
                    if np.take(packet.gw_data['gw_clock'], i) not in self.gw_data['gw_clock'] or \
                            np.take(packet.gw_data['time_from_start'], i) not in self.gw_data['time_from_start']:
                        for key in gw_data_info.keys():
                            self.gw_data[key] = np.append(self.gw_data[key], np.take(packet.gw_data[key], i))
                        for key in self.custom_data.keys():
                            if key in packet.custom_data.keys():
                                self.custom_data[key] = np.append(self.custom_data[key],
                                                                  np.take(packet.custom_data[key], i))
                            else:
                                self.custom_data[key] = np.append(self.custom_data[key], [None])
                        if do_sort:
                            sort_idx = self.gw_data['time_from_start'].argsort()
                            for k, v in self.gw_data.items():
                                self.gw_data[k] = v[sort_idx]
                            for k, v in self.custom_data.items():
                                self.custom_data[k] = v[sort_idx]
                    else:
                        if output_log:
                            self.printing_and_logging('Tried to add duplicated packet to sprinkler {}'.format(packet))
                except Exception as e:
                    self.printing_and_logging('Failed to add packet {} to sprinkler, exception: {}'
                                              .format(packet, str(e)))
        else:
            self.printing_and_logging('Not from the same sprinkler')
            status = False
        # self.sort()
        return status

    def as_dict(self, sprinkler_index=None):  # None not tested
        packet_data = self.packet_data.copy()
        sprinkler_gw_data = self.gw_data.copy()
        custom_data = self.custom_data.copy()
        if sprinkler_index is not None:
            if sprinkler_index > self.gw_data['stat_param'].size:
                return None
            for gw_attr in gw_data_info.keys():
                sprinkler_gw_data[gw_attr] = np.take(self.gw_data[gw_attr], sprinkler_index)
            for custom_attr in self.custom_data.keys():
                custom_data[custom_attr] = np.take(self.custom_data[custom_attr], sprinkler_index)
            dict_len = 1
        else:
            dict_len = len(self)
            for k, v in packet_data.items():
                if k == 'raw_packet' and 'adi_per_packet' in custom_data.keys():
                    adi_per_packet_list = custom_data['adi_per_packet']
                    if not isinstance(custom_data['adi_per_packet'], list) and \
                            not isinstance(custom_data['adi_per_packet'], np.ndarray):
                        adi_per_packet_list = [adi_per_packet_list] * dict_len
                    original_raw_packets = [construct_packet_from_fields(
                        packet_data=self.packet_data,
                        fields_to_convert={'adi': adi_per_packet}) for adi_per_packet in adi_per_packet_list]
                    packet_data[k] = original_raw_packets
                    continue
                packet_data[k] = [v] * dict_len

        data = {**packet_data, **sprinkler_gw_data}
        data['gw_process'] = [self.gw_process] * dict_len  # bool should be a list for pd.DataFrame.from_dict(data)
        data['is_valid_packet'] = [self.is_valid_packet] * dict_len  # bool should be a list
        data['inlay_type'] = [self.inlay_type] * dict_len
        for k, v in custom_data.items():  # add to df only custom data keys that are not part of the packet
            if k not in data:
                data[k] = v
        return data

    def as_dataframe(self, sprinkler_index=None):
        data = self.as_dict(sprinkler_index=sprinkler_index)
        packet_df = pd.DataFrame.from_dict(data)

        return packet_df

    def get_per(self, expected_sprinkler_count=None):
        """
        Calculates the packet per at the sprinkler
        @param expected_sprinkler_count - The number of packet per sprinkler
        @return packet per at percentage
        """
        if expected_sprinkler_count is None:
            if self.packet_data['test_mode']:
                expected_sprinkler_count = 6
            else:
                expected_sprinkler_count = get_flow_param(flow_in=self.packet_data['flow_ver'],
                                                          param_name='sprinkler_num')
        return 1 - (len(self) / expected_sprinkler_count)

    def get_tbp(self):
        """
        calculates the rate of packets from the same sprinkler
        :return: min_times_found - in msec
        :rtype: int
        """

        def triad_ratio_logic(diff_time_1, diff_time_2, ratio=1.0, error=10):
            """ estimate the time between successive packet according to only 3 packets out of 6 """
            if abs(diff_time_1 - ratio * diff_time_2) <= diff_time_2 / error:
                return True
            elif abs(diff_time_1 - (1 / ratio) * diff_time_2) <= diff_time_1 / error:
                return True
            else:
                return False

        def estimate_diff_packet_time(times_list, pc_time_list):
            if times_list.size < sprinkler_num / 2 or times_list.size == 1:
                return None

            # sorting based on the pc time:
            sort_idx = pc_time_list.argsort()
            pc_time_list_sorted = pc_time_list.copy()[sort_idx]
            times_list_sorted = times_list.copy()[sort_idx]

            dt = []
            for i, t_hw in enumerate(zip(times_list_sorted[1::], times_list_sorted[:-1])):
                dt_tmp = t_hw[0] - t_hw[1]
                dt_pc = round((pc_time_list_sorted[i + 1] - pc_time_list_sorted[i]) * 1000)
                if dt_pc >= MAX_GW_CLOCK:
                    # HW timing was zeroing during the same packets sprinkler
                    dt_tmp = dt_pc
                elif dt_tmp < 0:
                    dt_tmp += MAX_GW_CLOCK

                if abs(dt_pc - dt_tmp) > MAX_TIME_DIFF * 1000:
                    dt_tmp = dt_pc

                dt.append(dt_tmp)

            return dt

        def check_if_nan(lst):
            is_nan = pd.isnull(lst)
            if isinstance(is_nan, bool):
                return is_nan
            if isinstance(is_nan.tolist(), bool):
                return is_nan.tolist()
            if isinstance(is_nan, np.ndarray):
                return any(is_nan)
            return False

        if check_if_nan(self.gw_data['time_from_start']) or check_if_nan(self.gw_data['gw_clock']):
            return None

        if self.packet_data['test_mode']:
            sprinkler_num = 6
        else:
            sprinkler_num = get_flow_param(flow_in=self.packet_data['flow_ver'], param_name='sprinkler_num')
        estimate_diff_time = estimate_diff_packet_time(self.gw_data['gw_clock'], self.gw_data['time_from_start'])
        if estimate_diff_time is None:
            return None
        elif len(self) == sprinkler_num/2 and sprinkler_num == 6:
            # gen 2 - got 3 packets, need to estimate
            if triad_ratio_logic(estimate_diff_time[0], estimate_diff_time[1], ratio=1):
                # equal distance between packets, can't decide if successive
                return None
            else:
                is_estimated = False
                for ratio in [2, 3, 4]:
                    if triad_ratio_logic(estimate_diff_time[0], estimate_diff_time[1], ratio=ratio):
                        estimate_diff_time = [min(estimate_diff_time[0], estimate_diff_time[1]),
                                              max(estimate_diff_time[0], estimate_diff_time[1]) / ratio]
                        is_estimated = True
                        break
                if not is_estimated and triad_ratio_logic(estimate_diff_time[0], estimate_diff_time[1], ratio=1.5):
                    estimate_diff_time = [min(estimate_diff_time[0], estimate_diff_time[1]) / 2,
                                          max(estimate_diff_time[0], estimate_diff_time[1]) / 3]
                    is_estimated = True
            if not is_estimated:
                return None

        return int(min(estimate_diff_time))

    def extract_packet_data_by_name(self, key):
        """
        extract data from all packet attribute according to the key name
        :param key: the name of the data (the key of the relevant dictionary)
        :type key: str
        :return: list of the data
        :rtype: list
        """

        if key in self.packet_data:
            data_attr = self.packet_data[key]
        elif hasattr(self, 'decoded_data') and key in self.decoded_data:
            data_attr = self.decoded_data[key]
        elif key in self.gw_data:
            data_attr = self.gw_data[key]
        elif key in self.custom_data:
            data_attr = self.custom_data[key]
        else:
            self.printing_and_logging('key:{} does not exist in packet structure')
            return None
        if isinstance(data_attr, np.ndarray):
            data_attr = data_attr.tolist()
        if not isinstance(data_attr, list):
            data_attr = [data_attr]
        return data_attr

    def filter_by_sprinkler_id(self, sprinkler_ids):
        """
        keep only specific sprinklers in a packet according to the sprinkler id
        :param sprinkler_ids:
        :type sprinkler_ids: list
        :return: filtered packet
        :rtype: Packet or DecryptedPacket
        """

        def filter_per_attr(data_attr):
            for key in data_attr.keys():
                if isinstance(data_attr[key], list) or \
                        (isinstance(data_attr[key], np.ndarray) and data_attr[key].size > 1):
                    filtered_data = [data_attr[key][i] for i in sprinkler_ids]
                    if isinstance(data_attr[key], np.ndarray):
                        data_attr[key] = np.array(filtered_data)
                    else:
                        data_attr[key] = filtered_data
            return data_attr

        filtered_packet = self.copy()
        filtered_packet.gw_data = filter_per_attr(filtered_packet.gw_data)
        filtered_packet.custom_data = filter_per_attr(filtered_packet.custom_data)

        return filtered_packet

    def add_custom_data(self, custom_data):
        for key in custom_data.keys():
            if isinstance(custom_data[key], list):
                if len(custom_data[key]) == self.__len__():
                    self.custom_data[key] = custom_data[key]
                else:
                    self.printing_and_logging('add_custom_data failed - '
                                              'the custom data is a list of a different length than'
                                              ' the number of packets')
            else:
                self.custom_data[key] = self.__len__() * [custom_data[key]]


if __name__ == '__main__':

    # test mode packets with different adi, same sprinkler:
    testing_82_tm_adi0 = Packet(
        'full_packet("4731090906D03E34CD8200001EFF000505000060750774415ACDFFFBD069756C550DB1FD75E7B127312C21F3B10A03B1CB12AF2D45AC")')
    testing_82_tm_adi1 = Packet(
        'full_packet("4731090906D03E34CD8200011EFF000505000060750774415ACDFFFBD069756C550DB1FD75E7B127312C21F3B10A03B1CB12AF2D45CE")')
    testing_82_tm_adi2 = Packet(
        'full_packet("4731090906D03E34CD8200031EFF000505000060750774415ACDFFFBD069756C550DB1FD75E7B127312C21F3B10A03B1CB12AF2D4615")')
    testing_82_tm_adi0.append_to_sprinkler(testing_82_tm_adi1)
    testing_82_tm_adi0.append_to_sprinkler(testing_82_tm_adi2)
    print(f'packet len: {len(testing_82_tm_adi0)} with the following adi: '
          f'{testing_82_tm_adi0.custom_data["adi_per_packet"]}')

    packet_1 = {'packet': '03B28DCD99201EFF0005FE0000E0210FFF93B635EBFF1DB118C6D782DC2ED98C404200486436AE8F',
                'is_valid_tag_packet': True, 'adv_address': '03B28DCD9920', 'group_id': 'FE0000', 'rssi': 54,
                'stat_param': 30687,
                'time_from_start': 1.528374}
    packet_2 = '03B28DCD99201EFF0005FE0000E0210FFF93B635EBFF1DB118C6D782DC2ED98C404200486436AE82'

    p1 = Packet(packet_1)

    print(p1.get_packet_content('03B28DCD99201EFF0005FE0000E0210FFF93B635EBFF1DB118C6D782DC2ED98C404200486436AE82', get_gw_data=True),)

    p2 = Packet(packet_2)

    p3 = Packet('0437E62F33341E16AFFD02000058AD2AE76DE1D314B003083386A2A1AD139E42C10A0448F93472DC')
    print(p3.get_payload())

    packet_no_gw = Packet('03B28DCD99201EFF0005FE0000E0210FFF93B635EBFF1DB118C6D782DC2ED98C4042004864')
    packet_corrupted_length = Packet('03B28DCD99201EFF0005FE0000E0210FFF93B635EBFF1DB118C6D782DC2ED98C4042004864x')

    print(p1.get_packet_string(0))
    print(p1.get_average_rssi())

    print(p1 == p2)
    print(p1.append_to_sprinkler(p2))

    p6_1 = Packet(raw_packet='473109090673E063E00800002616AFFDFE0080FB7140E0F9B527E8E4751A0A18A99E360281421735AF8A000281415D687CC27F529230',
                  time_from_start=81.092617)
    p6_2 = Packet(raw_packet='473109090673E063E00800002616AFFDFE0080FB7140E0F9B527E8E4751A0A18A99E360281421735AF8A000281415D687CC27F52DEA2',
                  time_from_start=228.187184)
    p6 = p6_1.append_to_sprinkler(p6_2)
    tbp = p6_1.get_tbp()

    p6_3 = Packet(raw_packet='473109090695564AAB0800002616AFFDFE00C05AA9955A56FF3B552A24B28D275BDC8647CFDA0750ABA8541FF99340948F41D450AFCC',
                  time_from_start=38.627151)
    p6_4 = Packet(
        raw_packet='473109090695564AAB0800002616AFFDFE00C05AA9955A56FF3B552A24B28D275BDC8647CFDA0750ABA8541FF99340948F41D451BB8E',
        time_from_start=105.55923)
    p6a = p6_3.append_to_sprinkler(p6_4)
    tbp2 = p6_3.get_tbp()

    p6_5 = Packet(
        raw_packet='473109090695564AAB0800002616AFFDFE00C05AA9955A56FF3B552A24B28D275BDC8647CFDA0750ABA8541FF99340948F41D450FFFC',
        time_from_start=0.0)
    p6_6 = Packet(
        raw_packet='473109090695564AAB0800002616AFFDFE00C05AA9955A56FF3B552A24B28D275BDC8647CFDA0750ABA8541FF99340948F41D4510010',
        time_from_start=10.55923)
    p6b = p6_5.append_to_sprinkler(p6_6)
    tbp3 = p6_5.get_tbp()

    p6_7 = Packet(
        raw_packet='06455AD22E061E16AFFDFE0040FFBD941223357780F3FBC45BA7E13D7E6AFB4F89D61463EA3EFF9A',
        time_from_start=55796.199652)
    p6_8 = Packet(
        raw_packet='06455AD22E061E16AFFDFE0040FFBD941223357780F3FBC45BA7E13D7E6AFB4F89D61463EA3E00DA',
        time_from_start=55796.199652)
    p67 = p6_7.append_to_sprinkler(p6_8)
    tbp4 = p6_7.get_tbp()

    print(len(p1))
    print(p1.get_average_rssi())

    p1_dict = p1.as_dict()
    p1_df = pd.DataFrame(data=p1_dict)
    print(p1_df)

    print('end')



"""
Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import logging
import numpy as np
import pandas as pd
import copy
import inspect
import re
import os

from wiliot_core.packet_data.packet import Packet, InlayTypes
from wiliot_core.packet_data.config_files.parse_packet import packet_length_types, extract_packet_data_types, extract_gw_data_types
from wiliot_core.packet_data.utils.convert_cloud_packets import ConvertCloudPacket
from wiliot_core.packet_data.config_files.stat_calculation import StatisticsCalc


class PacketList(list):
    def __init__(self, packet_obj_type=Packet, packet_list_obj_type=None, list_custom_data=None, logger_name=None):
        """
        :param packet_obj_type:
        :param packet_list_obj_type:
        :param list_custom_data: the packet_list custom data as dictionary
        :type list_custom_data: dict
        :param logger_name: the log file we want to write into it.
        """
        self.packet_list = np.array([], dtype=Packet)  # ndarray - contains packets
        self.raw_packet_map_list = {}  # dict - keys: raw_packet, values: index in self.packet_list
        self.is_df_changed = True
        self.packet_df = None
        self.packet_obj_type = packet_obj_type
        if packet_list_obj_type is None:
            self.packet_list_obj_type = PacketList
        else:
            self.packet_list_obj_type = packet_list_obj_type

        if list_custom_data is not None:
            self.list_custom_data = list_custom_data
        else:
            self.list_custom_data = {}
        if logger_name is not None:
            self.logger = logging.getLogger(logger_name)
        else:
            self.logger = None

    def __add__(self, other_packet_list):
        """
        merge 2 PacketList object using '+' sign: packet_list1+packet_list2

        :type other_packet_list: PacketList
        :param other_packet_list:

        :return: merged packet_list, not mandatory
        """
        for packet in other_packet_list.packet_list:
            self.append(packet=packet)

        return self

    def printing_and_logging(self, message):
        if self.logger:
            self.logger.warning(message)
        else:
            print(message)

    def __len__(self):
        """
        Total amount of packets (sprinklers calculated separately)
        """
        return self.size()

    def __iter__(self):
        self.n = 0
        return iter(self.packet_list)

    def __next__(self):
        if self.n <= len(self):
            return self.packet_list[self.n]
        else:
            raise StopIteration

    def __getitem__(self, key):
        return self.packet_list[key]

    def __setitem__(self, key, packet: Packet):
        # raise ValueError("Set item is not supported yet by PacketList type")
        self.is_df_changed = True
        if isinstance(packet, Packet):
            old_packet = self.packet_list[key]
            del self.raw_packet_map_list[old_packet.get_packet_string(gw_data=False, process_packet=False)]
            self.raw_packet_map_list[packet.get_packet_string(gw_data=False, process_packet=False)] = key
            self.packet_list[key] = packet
        else:
            raise TypeError("Can only set Packet type to PacketList")

    def pop(self, index=0):
        """

        :param index: index to pop out of the list
        :type index: int
        :return: the shorter packet list
        :rtype: PacketList
        """
        packet_list = self.copy()
        raw_packet = np.take(packet_list.packet_list, index).packet_data['raw_packet']
        packet_list.packet_list = np.delete(packet_list.packet_list, index)
        packet_list.raw_packet_map_list.pop(raw_packet, None)
        return packet_list

    def copy(self):
        if self.size() == 0:
            return copy.deepcopy(self)
        else:
            copy_packet_list = self.packet_list_obj_type()
            for p in self.packet_list:
                copy_packet_list.append(p.copy())
            return copy_packet_list

    def size(self):
        """
        Total amount of packets, sprinklers count individually
        """
        packet_list_size = 0
        for packet in self.packet_list:
            packet_list_size += len(packet)
        return packet_list_size

    def append(self, packet: Packet, ignore_sprinkler=False, sort_by_time=False):
        """
        Adds single Packet to PacketList

        :type packet: Packet
        :param packet: packet to be added to packet_list
        :type ignore_sprinkler: Bool
        :param ignore_sprinkler: allow duplicates packets from different sprinkler
        :type sort_by_time: bool
        :param sort_by_time: if true append to sprinkler based on the time from start

        :return: packet_list
        """
        if packet.is_valid_packet:
            self.is_df_changed = True
            raw_packet = packet.get_packet()
            if raw_packet not in self.raw_packet_map_list.keys():
                self.packet_list = np.append(self.packet_list, packet)
                raw_packet_list_index = self.packet_list.size - 1
                self.raw_packet_map_list[raw_packet] = raw_packet_list_index
            else:
                raw_packet_list_index = self.raw_packet_map_list[raw_packet]
                self.packet_list[raw_packet_list_index].append_to_sprinkler(packet=packet, do_sort=sort_by_time)

    def get_sprinkler(self, packet):
        """
        @param packet: a packet data type. Needs to be appended before using this function.
        @return: A packet sprinkler of this packet.
        """
        raw_packet = packet.get_packet()
        if raw_packet in self.raw_packet_map_list.keys():
            raw_packet_list_index = self.raw_packet_map_list[raw_packet]
            return self.packet_list[raw_packet_list_index]
        else:
            return None

    def dump(self, packet_dict_list: list):
        """
        gets list of raw_packet or packet_dict and fill packet_list with data

        :type packet_dict_list: list
        :param packet_dict_list: gw list (get_packets), fill packet_list with data

        :return: bool status
        """
        if len(self) > 0:
            # packet_list not empty
            return False
        self.is_df_changed = True
        for packet_dict in packet_dict_list:
            packet = self.packet_obj_type(packet_dict,
                                          logger_name=self.logger.name if self.logger is not None else None)
            if packet.is_valid_packet:
                self.append(packet)
        return True

    def update_packet_df(self, packet_df):
        self.packet_df = packet_df
        self.is_df_changed = False

    def get_df(self, sprinkler_filter=False, add_sprinkler_info=True, is_overwrite=False, as_dict=False,
               first_packet_id=0):
        """
        returns packet_list data as dataframe, BE CAREFULL WHEN CHANGING FUNCTIONALITY, this is a fundamental function

        :type sprinkler_filter: Bool
        :param sprinkler_filter: determine if to keep all occurrences of sprinkler per packet
        :type add_sprinkler_info: Bool
        :param add_sprinkler_info: adds 'num_packets_per_sprinkler','per','tbp' attributes to df
        :param is_overwrite: if True, get df overwrite columns with the same name
        :type is_overwrite: bool
        :param as_dict: if True, returns a dictionary instead of dataframe
        :type as_dict: bool
        :param first_packet_id: if specified, the df include packets only from this packet id and larger
        :type first_packet_id: int
        :return: Dataframe
        """
        if not self.is_df_changed:
            sprinkler_info_calculated = 'packet_id' in self.packet_df.keys()
            if sprinkler_filter and sprinkler_info_calculated:
                return self.packet_df.drop_duplicates(subset='packet_id')
            elif (sprinkler_info_calculated and not is_overwrite) or not add_sprinkler_info:
                return self.packet_df

        all_packets_dict = []
        for packet_id, packet in enumerate(self.packet_list):
            if packet_id < first_packet_id:
                continue
            sprinkler_dict = packet.as_dict()
            # add sprinkler info:
            if sprinkler_filter is True or add_sprinkler_info is True:
                tbp = packet.get_tbp()
                if tbp is None:
                    tbp = -1
                per = packet.get_per()

                sprinkler_keys = ['packet_id', 'sprinkler_counter', 'num_packets_per_sprinkler', 'tbp', 'per']
                sprinkler_values = [[packet_id] * len(packet), list(range(1, len(packet) + 1)),
                                    [len(packet)] * len(packet), [tbp] * len(packet), [per] * len(packet)]
                for k, v in zip(sprinkler_keys, sprinkler_values):
                    if k in sprinkler_dict.keys() and not is_overwrite:
                        continue
                    sprinkler_dict[k] = v
            # append packet dict:
            if isinstance(sprinkler_dict['raw_packet'], list) and len(sprinkler_dict['raw_packet']) > 1:
                for ind in range(len(sprinkler_dict['raw_packet'])):
                    all_packets_dict.append({k: v[ind] for k, v in sprinkler_dict.items()})
            else:
                for k, v in sprinkler_dict.items():
                    if isinstance(v, list):
                        sprinkler_dict[k] = v[0]
                    elif isinstance(v, np.ndarray):
                        sprinkler_dict[k] = v.item()
                all_packets_dict.append(sprinkler_dict)

        if as_dict:
            return all_packets_dict

        self.packet_df = pd.DataFrame(all_packets_dict)
        self.is_df_changed = False

        if sprinkler_filter:
            if len(self.packet_df['packet_id'].unique()) == 1:
                return self.packet_df.head(1)
            return self.packet_df.drop_duplicates(subset='packet_id')
        else:
            return self.packet_df

    def packet_df_to_sprinkler_df(self, packet_df):
        """
        gets packet_df and returns sprinkler df
        :type packet_df: DataFrame
        :param packet_df: df to convert
        :return: Dataframe - sprinkler_df
        """
        packet_list = self.import_packet_df(packet_df=packet_df)
        sprinkler_df = packet_list.get_df(sprinkler_filter=True, add_sprinkler_info=True)

        return sprinkler_df

    def sort_df_by(self, column='gw_clock'):
        """
        returns dataframe sorted by column
        :type column: str
        :param column: the column to filter by

        :return: Dataframe
        """
        packet_df = self.get_df()
        packet_df_sorted = packet_df.sort_values(column)
        return packet_df_sorted

    def filter_packet_by(self, packet_data_key='adv_address', values='', filter_out=False):
        """
        filter packet_list by adv_address or any other data key
        :type packet_data_key: str
        :param packet_data_key:
        :type values: str or list of strings
        :param values: e.g. adv_address to search
        :type filter_out: bool
        :param filter_out: if True return all packets that does NOT have packet_data_key values

        :return: filtered PacketList
        """
        packet_list = self.packet_list_obj_type(list_custom_data=self.list_custom_data)
        for index, packet in enumerate(self.packet_list):
            filter_data = packet.extract_packet_data_by_name(packet_data_key)
            if filter_data is None:
                p_valid = 'valid' if packet.is_valid_packet else 'invalid'
                self.printing_and_logging(f'PacketList->filter_packet_by: packet_data_key does not exist '
                                          f'for {p_valid} packet:{packet.get_packet_string()}')
                continue

            if not isinstance(values, list):
                values = [values]

            if len(filter_data) > 1:  # filter by sprinklers
                filtered_sprinkler_id = [i for i, f_d in enumerate(filter_data) if (not filter_out and f_d in values) or
                                         (filter_out and f_d not in values)]
                if len(filtered_sprinkler_id):
                    packet_list.append(self.packet_list[index].filter_by_sprinkler_id(filtered_sprinkler_id))
            else:  # filter by packet
                if (not filter_out and filter_data[0] in values) or (filter_out and filter_data[0] not in values):
                    packet_list.append(self.packet_list[index])

        return packet_list

    def filter_df_by(self, packet_df=None, column='adv_address', values='', values_range=[]):
        """
        filter dataframe by value or (exclusive) by value range
        :type packet_df: DataFrame
        :param packet_df: gets get_df dataframe (packet_df only)
        :type column: str
        :param column: column to filter by
        :type values: str or int or list of strings
        :param values: value to search
        :type values_range: 2 elements list of int
        :param values_range: range to find

        :return: filtered Dataframe
        """
        if packet_df is None:
            packet_df = self.get_df()
        if values_range:
            start = values_range[0]
            end = values_range[1]
            packet_df_filtered = packet_df.loc[(packet_df[column] > start) & (packet_df[column] <= end)]
        else:
            if isinstance(values, list):
                packet_df_filtered = packet_df.loc[packet_df[column].isin(values)]
            else:
                packet_df_filtered = packet_df.loc[packet_df[column] == values]
        return packet_df_filtered

    def get_avg_rssi(self):
        """
        return packet list average rssi (4 decimal points accuracy)
        :return: average rssi for tag
        """
        all_rssi = np.array([])
        for packet in self.packet_list:
            if 'rssi' not in packet.gw_data.keys():
                continue
            for i in range(len(packet)):
                all_rssi = np.append(all_rssi, np.take(packet.gw_data['rssi'], i))

        if len(all_rssi) == 0:
            return float('nan')
        avg_rssi = round(all_rssi.mean(), 4)
        return avg_rssi

    def get_avg_tbp(self, ignore_outliers=False):
        """
        return packet list average tbp (4 decimal points accuracy)
        :param ignore_outliers: reject data outside the 2 std area
        :type ignore_outliers: bool

        :return: average tbp for tag
        """

        def reject_outliers(data, m=2):
            return data[abs(data - np.mean(data)) < m * np.std(data)]

        tbp_list = np.array([])
        for packet in self.packet_list:
            packet_tbp = packet.get_tbp()
            if packet_tbp is not None:
                tbp_list = np.append(tbp_list, packet_tbp)
        if len(tbp_list) == 0:
            return None
        avg_tbp = round(tbp_list.mean(), 4)
        if ignore_outliers:
            avg_tbp = round(reject_outliers(tbp_list).mean(), 4)
        return avg_tbp

    def to_csv(self, path, append=False, export_packet_id=True, columns=None, add_sprinkler_info=True,
               is_overwrite=False):
        """
        export entire PacketList to csv

        :type path: str
        :param path: path to save csv
        :type append: bool
        :param append: to append the df to an existing csv file

        :return: bool - export status
        """

        return self.export_packet_df(packet_df=self.get_df(add_sprinkler_info=add_sprinkler_info,
                                                           is_overwrite=is_overwrite),
                                     path=path,
                                     append=append,
                                     export_packet_id=export_packet_id, columns=columns)

    def export_packet_df(self, packet_df, path, append=False, export_packet_id=True, columns=None):
        """
        export given dataframe to csv

        :type packet_df: Dataframe
        :param packet_df: filtered dataframe to save as csv
        :type path: str
        :param path: path to save csv
        :type append: bool
        :param append: to append the df to an existing csv file

        :return: bool - export status
        """
        try:
            # this call of 'to_csv' is a generic pandas function
            if not export_packet_id and 'packet_id' in packet_df:
                packet_df.drop('packet_id', axis=1, inplace=True)

            if append:
                packet_df.to_csv(path, mode='a', index=False, header=False, columns=columns)
            else:
                packet_df.to_csv(path, index=False, columns=columns)
            return True
        except KeyError as e:

            raise KeyError(f'problem during export packet df {e}')

    def packet_string_to_object(self, packet_in, time_from_start, configs=None, custom_data=None, inlay_type=None,
                                ignore_crc=False):
        packet_obj = Packet(packet_in, float(time_from_start), custom_data=custom_data, inlay_type=inlay_type,
                            logger_name=self.logger.name if self.logger is not None else None,
                            ignore_crc=ignore_crc)
        return packet_obj

    def import_packet_df(self, path=None, packet_df=None, custom_data_attr=None, import_all=False,
                         list_custom_data=None, verbose=True, inlay_type=None, ignore_crc=True,
                         obj_out=None):
        """
        import from a csv of dataframe
        :type path: str
        :param path: the message or part of the message that needed to be read
        :type packet_df: DataFrame
        :param packet_df: gets get_df dataframe (packet_df only)
        :type custom_data_attr: list
        :param custom_data_attr: if specified, only the specified additional coulmns in the attached df/csv will be
                                added to the packets as custom_data
        :type import_all: bool
        :param import_all: if True, all other columns in the attached df/csv will be added to the packets as custom_data
        :type list_custom_data: dict
        :param list_custom_data: the packet_list custom data as dictionary
        :type verbose: bool
        :param verbose: if False, remove some of the printing while converting each packet to packet object
        :type inlay_type: InlayTypes
        :param inlay_type: the packet_list inlay type
        :type ignore_crc: bool
        :param ignore_crc: if True, update for all packets crc_valid = True,
                           relevant for data collected by old GW FW version
        :type obj_out: PacketList or DecryptedPacketList or TagCollection or DecryptedTagCollection
        :param obj_out: if specified the return object from the function will be as the obj_out,
                        otherwise would be same as self object type
        :return: obj_out
        """
        class UnsupportedFileFormat(Exception):
            def __init__(self):
                supported_content_str = 'dataframe/file must contains one of the following columns conventions:\n' \
                    '1. raw_packet AND gw_packet\n2. encryptedPacket\n3. decrypted_full_packet\n4. encrypted_packet\n' \
                        '5. rawPacket AND packetVersion (tables from databricks based on cloud conversion that rawPacket is the packet payload\n'
                super().__init__(f'{supported_content_str}')
        non_packets_rows = []
        list_custom_data = list_custom_data if list_custom_data is not None else self.list_custom_data
        if obj_out is None:
            obj_out = self.packet_list_obj_type(list_custom_data=list_custom_data)
        if path is not None:
            import_packets_df = pd.read_csv(path)
        else:
            import_packets_df = packet_df

        first_timestamp = None
        static_keys_list = self.get_generic_df_cols() if import_all else None
        for index, row in import_packets_df.iterrows():
            try:
                reconstructed_packet = None
                if 'raw_packet' in row.keys():
                    raw_packet = row['raw_packet']
                    gw_data = row['gw_packet'] if 'gw_packet' in row.keys() else ''
                    if not pd.isnull(raw_packet) and not pd.isnull(gw_data):
                        reconstructed_packet = str(raw_packet) + str(gw_data)
                elif 'encryptedPacket' in row.keys():
                    reconstructed_packet = row['encryptedPacket']
                elif 'decrypted_full_packet' in row.keys():
                    reconstructed_packet = row['decrypted_full_packet']
                elif 'encrypted_packet' in row.keys():
                    reconstructed_packet = row['encrypted_packet']
                elif 'rawPacket' in row.keys() and ('packetVersion' in row.keys() or 'flowVersion' in row.keys()):
                    if first_timestamp is None:
                        if 'timestamp' in import_packets_df.keys():
                            first_timestamp = import_packets_df['timestamp'].min()
                        else:
                            first_timestamp = 0
                    c = ConvertCloudPacket(enriched_packet_row=row, first_timestamp=first_timestamp)
                    reconstructed_packet = c.get_reconstruct_packet()
                else:
                    raise UnsupportedFileFormat()
                if reconstructed_packet == '' or pd.isnull(reconstructed_packet):
                    non_packets_rows.append(index)
                    continue
            except UnsupportedFileFormat:
                raise
            except Exception as e:
                non_packets_rows.append(index)
                continue

            if 'time_from_start' in row.keys():
                time_from_start = row['time_from_start']
            elif 'time' in row.keys() and (not isinstance(row['time'], str) or
                                           row['time'].replace('.', '').isnumeric()):
                time_from_start = row['time']
            elif 'timestamp' in row.keys() and (
                    not isinstance(row['timestamp'], str) or row['timestamp'].replace('.', '').isnumeric()):
                time_from_start = float(row['timestamp']) - (first_timestamp if first_timestamp is not None else 0)
                time_from_start = float(time_from_start / 1000)  # timestamp is usually in ms
            else:
                time_from_start = float('nan')

            try:
                time_from_start = float(time_from_start)
            except ValueError:
                time_from_start = float('nan')

            # adding custom data
            custom_data = {}
            try:
                if import_all:
                    for k in row.keys():
                        if k not in static_keys_list:
                            custom_data[k] = row[k]
                elif custom_data_attr is not None:
                    for k in custom_data_attr:
                        if k in row.keys():
                            custom_data[k] = row[k]
            except Exception as e:
                    self.printing_and_logging(f'exception raised during extracting custom data in import df: {e}')
            try:
                p = self.packet_string_to_object(packet_in=reconstructed_packet,
                                                 time_from_start=time_from_start,
                                                 configs=row,
                                                 custom_data=custom_data,
                                                 inlay_type=inlay_type,
                                                 ignore_crc=ignore_crc)
                obj_out.append(p)
            except Exception as e:
                self.printing_and_logging('exception raised during import df: {}'.format(e))

        if len(non_packets_rows) and verbose:
            self.printing_and_logging('the following {} rows could not be imported '
                                      'to the packet list:{}'.format(len(non_packets_rows), non_packets_rows))
        return obj_out

    def get_statistics(self, packet_df=None):
        """
        Calculates statistics of self.
        @return dictionary with predefined statistics of the packetList.
        """
        return self.get_df_statistics(packet_df=packet_df)

    def get_group_statistics(self, group_by_col='adv_address'):
        """
        Calculates statistics of self, grouped by some value.
        @return dictionary of items at the group, Dictionary values are dictionaries of statistics.
        """
        packet_df = self.get_df()
        # if not defines - get all values:

        groups_id_list = packet_df[group_by_col].unique()

        group_statistics = {}
        for group_id in groups_id_list:
            packet_df_filtered = packet_df.loc[packet_df[group_by_col] == group_id]
            group_statistics[group_id] = self.get_df_statistics(packet_df=packet_df_filtered)
            group_statistics[group_id][group_by_col] = group_id
        return group_statistics

    def group_by(self, group_by_col='adv_address'):
        """
        Calculates statistics of self, grouped by some value.
        @return dictionary of items at the group, Dictionary values are dictionaries of statistics.
        """
        packet_df = self.get_df(sprinkler_filter=True)
        # if not defines - get all values:

        groups_id_list = packet_df[group_by_col].unique()

        group = {}
        for group_id in groups_id_list:
            group[group_id] = self.filter_packet_by(group_by_col, group_id)
            # group[group_id][group_by_col] = group_id
        return group

    def get_num_packets(self):
        """
        Total amount of packets, sprinklers packets count separately
        """
        return len(self.packet_list)

    def get_df_statistics(self, packet_df=None, stat_calc=StatisticsCalc):
        """
        Calculates statistics from packetList DF.
        @param packet_df - dataframe generated by packetList
        @param stat_calc - object to calc the statistics
        @return dictionary with predefined statistics of the packetList.
        """
        statistics = {}
        if (packet_df is not None and packet_df.shape[0] == 0) or (packet_df is None and self.size() == 0):
            statistics = {'num_packets': 0, 'num_cycles': 0}
            return statistics

        # sprinkler_df = self.get_df(sprinkler_filter=True)
        if packet_df is None:
            packet_df = self.get_df(add_sprinkler_info=True)
        if 'sprinkler_counter' in packet_df.keys():
            sprinkler_df = packet_df[packet_df['sprinkler_counter'] == 1]
        else:
            sprinkler_df = self.packet_df_to_sprinkler_df(packet_df)

        s = stat_calc(packet_df, sprinkler_df)
        stat_attr = (getattr(s, name) for name in dir(s) if not name.startswith('_'))
        stat_methods = [a for a in stat_attr if inspect.ismethod(a)]
        statistics = {}
        for method in stat_methods:
            calc_stat = method()
            statistics = {**statistics, **calc_stat}

        return statistics

    def to_tag_list(self):
        # import here to avoid infinte loop (packet_list calls tag_list and tags_list calls packet_list)
        from wiliot_core.packet_data.tag_collection import TagCollection
        tag_list = TagCollection()
        for packet in self.packet_list:
            tag_list.append(packet)
        return tag_list

    def sync_packet_list(self, new_packet_list, time_sync_new, my_source_name='original', new_source_name='new',
                         do_sort=False):
        """
        This function merge one packet list to the self packet list. if time_sync_list is specified it sync the data based on
        the time sync list
        """
        if pd.isnull(time_sync_new):
            raise ValueError(f'cannot sync packet list due to time sync new id null: {time_sync_new}')

        for p in self.packet_list:
            p.add_custom_data({'source': my_source_name})

        for p in new_packet_list:
            p.gw_data['time_from_start'] = np.array(p.gw_data['time_from_start'] - time_sync_new)
            p.add_custom_data({'source': new_source_name, 'timestamp': float('nan')})
            self.append(p, sort_by_time=do_sort)

        if do_sort:
            min_time_from_start = []
            for p in self.packet_list:
                min_time_from_start.append(np.min(p.gw_data['time_from_start']))

            min_time_from_start = np.array(min_time_from_start)
            idx = min_time_from_start.argsort()
            self.packet_list = self.packet_list[idx]

    def get_generic_df_cols(self):
        packet_data = list(extract_packet_data_types().keys())
        gw_data = list(extract_gw_data_types().keys())
        return packet_data + gw_data
    
    def log_file_to_packet_struct(self, log_path, ignore_crc=False, packet_obj=None, packet_list_obj=None):
        packet_obj = packet_obj if packet_obj is not None else Packet
        packet_list_obj = packet_list_obj if packet_list_obj is not None else PacketList
        try:
            packet_list_out = packet_list_obj()
            user_event_out = []
            if os.path.isfile(log_path):
                f = open(log_path, 'r')
                lines = f.readlines()

                for i, line in enumerate(lines):
                    try:
                        if 'packet' in line:
                            # a data line
                            line_to_compare = line.replace(' ', '').replace("'", '')
                            if "raw:process_packet" in line_to_compare:
                                re_match = re.search(r"process_packet\(\"(\w+)\"", line)
                                packet_raw = str(re_match.groups(1)[0])
                            elif "packet:" in line_to_compare:  # packet: ABCD format (data type = processed
                                re_match = re.search(r",packet:(\w+)", line)
                                packet_raw = str(re_match.groups(1)[0])
                            elif "raw:full_packet" in line_to_compare:  # packet: ABCD format (data type = processed
                                re_match = re.search(r"full_packet\(\"(\w+)\"", line)
                                packet_raw = str(re_match.groups(1)[0])
                            else:
                                continue  # line without packets
                            if "'time': " in line:
                                re_match = re.search(r"'time': (\d+.\d+)", line)
                                packet_time = float(re_match.groups(1)[0])
                            else:
                                re_match = re.search(r"time:(\d+.\d+)", line)
                                packet_time = float(re_match.groups(1)[0])

                            packet_list_out.append(packet_obj(raw_packet=packet_raw, time_from_start=packet_time,
                                                              ignore_crc=ignore_crc))
                        if 'user_event' in line:
                            user_event_time = float('nan')
                            user_event_data = ''
                            if "user_event_time" in line:
                                re_match = re.search(r"user_event_time: (\d+.\d+),", line)
                                user_event_time = float(re_match.groups(1)[0])
                            if 'User event' in line:
                                user_event_data = line.split('User event: ')[-1].replace('\n', '')
                            user_event_out.append(
                                {'user_event_time': user_event_time, 'user_event_data': user_event_data})

                        if i % 1000 == 0:
                            print(f'analyzing row number {i}')
                    except Exception as e:
                        print(f'line {line} was failed to upload due to: {e}')
                f.close()
            if len(user_event_out) > 0:
                user_event_df = pd.DataFrame(user_event_out)
                user_event_df.to_csv(log_path.replace('.log', '_user_event.csv'), index=False)
            return packet_list_out
        except Exception as e:
            raise Exception(f'export packets from log was failed due to: {e}')


if __name__ == '__main__':
    import os
    p = PacketList()
    generic_df_cols = p.get_generic_df_cols()
    print(generic_df_cols)
    try:
        df_path = r"MY_DATA_FILE.csv"

        new_list = PacketList()
        new_list.append(Packet(
            'process_packet("04A3E93988271E16AFFD0200004E9D46293BEED581709F1214C88F6F3978E2136EB48E413533807C")'))
        new_list.get_df()

        new_list = new_list.import_packet_df(path=df_path, import_all=True)
        print(f'avg rssi: {new_list.get_avg_rssi()}')
        f_out = new_list.filter_packet_by(packet_data_key='rssi', values=[40, 60])
        new_list_df = new_list.get_df(sprinkler_filter=True, add_sprinkler_info=True, is_overwrite=True)
        new_list.to_csv(df_path.replace('.csv', '_decrypted.csv'))
        new_list_short = PacketList()
        new_list_short = new_list_short.import_packet_df(path=df_path, custom_data_attr=['common_run_name',
                                                                                         'external_id'])
        new_list_short_df = new_list_short.get_df(add_sprinkler_info=True)

    except Exception as e:
        print(e)

    print('done')



from enum import Enum


class EnergyPatterns(Enum):
    energy_pattern_17 = {'energy_pattern_idx': 'ENERGY_PATTERN_17', 'calibration_type': 'CALIBRATION_STANDARD', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons only, send 3 beacons on channel 37, 38 and 39\\r\\n'}
    energy_pattern_18 = {'energy_pattern_idx': 'ENERGY_PATTERN_18', 'calibration_type': 'CALIBRATION_STANDARD', 'num_of_energy_channels': '1', 'energy_channels': ['CHANNEL_39_FREQ'], 'sub1g_energy': False, 'desc': 'Energizing on ch39 preceding a beacon channel 37 38 and 39\\r\\n'}
    energy_pattern_24 = {'energy_pattern_idx': 'ENERGY_PATTERN_24', 'calibration_type': 'CALIBRATION_STANDARD', 'num_of_energy_channels': '1', 'energy_channels': ['CHANNEL_37_FREQ'], 'sub1g_energy': False, 'desc': 'Energizing on ch37 preceding a beacon channel 37 38 and 39\\r\\n'}
    energy_pattern_25 = {'energy_pattern_idx': 'ENERGY_PATTERN_25', 'calibration_type': 'CALIBRATION_STANDARD', 'num_of_energy_channels': '1', 'energy_channels': ['CHANNEL_38_FREQ'], 'sub1g_energy': False, 'desc': 'Energizing on ch38 preceding a beacon channel 37 38 and 39\\r\\n'}
    energy_pattern_26 = {'energy_pattern_idx': 'ENERGY_PATTERN_26', 'calibration_type': 'CALIBRATION_STANDARD', 'num_of_energy_channels': '1', 'energy_channels': ['2454'], 'sub1g_energy': False, 'desc': 'Energizing on 2454 preceding a beacon channel 37 38 and 39\\r\\n'}
    energy_pattern_27 = {'energy_pattern_idx': 'ENERGY_PATTERN_27', 'calibration_type': 'CALIBRATION_TEST', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons only, send 4 beacons on channels 37,38,39,39 2ms apart\\r\\n'}
    energy_pattern_29 = {'energy_pattern_idx': 'ENERGY_PATTERN_29', 'calibration_type': 'CALIBRATION_STANDARD', 'num_of_energy_channels': '3', 'energy_channels': ['CHANNEL_38_FREQ', '2454', 'CHANNEL_39_FREQ'], 'sub1g_energy': False, 'desc': 'Evenely energizing on 2426, 2454 and 2480, preceding a beacon channel 37 38 and 39\\r\\n'}
    energy_pattern_33 = {'energy_pattern_idx': 'ENERGY_PATTERN_33', 'calibration_type': 'CALIBRATION_SHIFTED', 'num_of_energy_channels': '1', 'energy_channels': ['2454'], 'sub1g_energy': False, 'desc': 'Shifed tranmission. send 3 beacons on channel 2403, 2427 and 2481, energizing on 2454\\r\\n'}
    energy_pattern_34 = {'energy_pattern_idx': 'ENERGY_PATTERN_34', 'calibration_type': 'CALIBRATION_SHIFTED', 'num_of_energy_channels': '1', 'energy_channels': ['2427'], 'sub1g_energy': False, 'desc': 'Shifed tranmission. send 3 beacons on channel 2403, 2427 and 2481, energizing on 2427\\r\\n'}
    energy_pattern_35 = {'energy_pattern_idx': 'ENERGY_PATTERN_35', 'calibration_type': 'CALIBRATION_SHIFTED', 'num_of_energy_channels': '1', 'energy_channels': ['2481'], 'sub1g_energy': False, 'desc': 'Shifed tranmission. send 3 beacons on channel 2403, 2427 and 2481, energizing on 2481\\r\\n'}
    energy_pattern_37 = {'energy_pattern_idx': 'ENERGY_PATTERN_37', 'calibration_type': 'CALIBRATION_EURO', 'num_of_energy_channels': '1', 'energy_channels': ['2450'], 'sub1g_energy': False, 'desc': 'Energizing in 2450MHz. Beacons (0dBm): 37 2428 39 2454 2467 2415 2441\\r\\n'}
    energy_pattern_39 = {'energy_pattern_idx': 'ENERGY_PATTERN_39', 'calibration_type': 'CALIBRATION_SKIP', 'num_of_energy_channels': '1', 'energy_channels': ['CHANNEL_27_FREQ'], 'sub1g_energy': False, 'desc': 'Energizing in 2460MHz\\r\\n'}
    energy_pattern_40 = {'energy_pattern_idx': 'ENERGY_PATTERN_40', 'calibration_type': 'CALIBRATION_EXIT_TM', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons only, send 4 beacons on channels 37,38,22,22 for 450us beacon duration\\r\\n'}
    energy_pattern_41 = {'energy_pattern_idx': 'ENERGY_PATTERN_41', 'calibration_type': 'MINI_RX', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons based on mini rx setting\\r\\n'}
    energy_pattern_42 = {'energy_pattern_idx': 'ENERGY_PATTERN_42', 'calibration_type': 'ENTER_TEST_MODE_01', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons to enter test mode channel 2406\\r\\n'}
    energy_pattern_43 = {'energy_pattern_idx': 'ENERGY_PATTERN_43', 'calibration_type': 'ENTER_TEST_MODE_07', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons to enter test mode channel 2418\\r\\n'}
    energy_pattern_44 = {'energy_pattern_idx': 'ENERGY_PATTERN_44', 'calibration_type': 'ENTER_TEST_MODE_27', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons to enter test mode channel 2460\\r\\n'}
    energy_pattern_47 = {'energy_pattern_idx': 'ENERGY_PATTERN_47', 'calibration_type': 'EXIT_TEST_MODE_01', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons to exit test mode channel 2406\\r\\n'}
    energy_pattern_48 = {'energy_pattern_idx': 'ENERGY_PATTERN_48', 'calibration_type': 'EXIT_TEST_MODE_07', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons to exit test mode channel 2418\\r\\n'}
    energy_pattern_49 = {'energy_pattern_idx': 'ENERGY_PATTERN_49', 'calibration_type': 'EXIT_TEST_MODE_27', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Beacons to exit test mode channel 2460\\r\\n'}
    energy_pattern_50 = {'energy_pattern_idx': 'ENERGY_PATTERN_50', 'calibration_type': 'CALIBRATION_STANDARD', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': True, 'desc': 'Energizing in 915Mhz, beacon train is generated right after the energy pattern\\r\\n '}
    energy_pattern_52 = {'energy_pattern_idx': 'ENERGY_PATTERN_52', 'calibration_type': 'CALIBRATION_SKIP', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': True, 'desc': 'Energizing in 915Mhz only\\r\\n'}
    energy_pattern_61 = {'energy_pattern_idx': 'ENERGY_PATTERN_61', 'calibration_type': 'CALIBRATION_EURO', 'num_of_energy_channels': '0', 'energy_channels': [], 'sub1g_energy': True, 'desc': 'Energizing in 917.5Mhz. Beacons: Neg12dBm: 37 38 39 0dBm: 37 2428 39 2454 2467 2415 2441\\r\\n'}
    energy_pattern_62 = {'energy_pattern_idx': 'ENERGY_PATTERN_62', 'calibration_type': 'CALIBRATION_EURO', 'num_of_energy_channels': '1', 'energy_channels': ['2450'], 'sub1g_energy': True, 'desc': 'Energizing in 2450MHz and Sub1g. Beacons (0dBm): 37 2428 39 2454 2467 2415 2441\\r\\n'}
    energy_pattern_99 = {'energy_pattern_idx': 'ENERGY_PATTERN_99', 'calibration_type': 'CALIBRATION_SKIP', 'num_of_energy_channels': '40', 'energy_channels': [], 'sub1g_energy': False, 'desc': 'Sweep over BLE channels with LBT\\r\\n'}



"""
Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import multiprocessing
import os
import csv
import sys
import subprocess

from appdirs import user_data_dir
import logging
import json
import pathlib
import re
import datetime
try:
    from tkinter import Tk, Label, Button, ttk
except ModuleNotFoundError as e:
    print(f'could not import tkinter: {e}')

PACKET_PREFIXES_MAPPING = ('process_packet', 'full_packet')


class QueueHandler(object):
    def __init__(self):
        self.is_mac = sys.platform == 'darwin'
        self.manager = multiprocessing.Manager() if self.is_mac else None

    def get_multiprocess_queue(self, queue_max_size):
        if self.is_mac:
            queue = self.manager.Queue(maxsize=queue_max_size)
        else:
            queue = multiprocessing.Queue(maxsize=queue_max_size)
        return queue


class WiliotDir:
    def __init__(self) -> None:
        self.local_appdata_dir = ''
        self.wiliot_root_path = ''
        self.common_dir_path = ''
        self.config_dir_path = ''
        self.user_config_path = ''
        self.tester_subdirectories = ['results', 'logs', 'configs']
        
        self.set_dir()
        self.create_dir(self.local_appdata_dir)
        self.create_dir(self.wiliot_root_path)
        self.create_dir(self.common_dir_path)
        self.create_dir(self.config_dir_path)
    
    def set_dir(self):
        try:
            if 'WILIOT_APP_ROOT_PATH' in os.environ.keys():
                print(os.environ['WILIOT_APP_ROOT_PATH'])
                self.wiliot_root_path = os.environ['WILIOT_APP_ROOT_PATH']
            else:
                self.local_appdata_dir = user_data_dir()
                self.wiliot_root_path = os.path.abspath(os.path.join(self.local_appdata_dir, 'wiliot'))
            
            self.common_dir_path = os.path.abspath(os.path.join(self.wiliot_root_path, 'common'))
            self.config_dir_path = os.path.abspath(os.path.join(self.common_dir_path, 'configs'))
            self.user_config_path = os.path.abspath(os.path.join(self.config_dir_path, 'user_configs.json'))
        
        except Exception as e:
            logging.warning('Error loading environment or getting in from OS, supporting Windows, Linux and MacOS '
                            '({})'.format(e))
    
    @staticmethod
    def create_dir(path):
        if not os.path.exists(path):
            os.makedirs(path)
    
    def create_tester_dir(self, tester_name):
        tester_path = self.get_tester_dir(tester_name)
        self.create_dir(tester_path)
        
        for subdir in self.tester_subdirectories:
            self.create_dir(tester_path + '/' + subdir)
    
    def get_tester_dir(self, tester_name):
        wiliot_path = self.wiliot_root_path
        tester_path = os.path.abspath(os.path.join(wiliot_path, tester_name))
        return tester_path
    
    def get_dir(self):
        return self.wiliot_root_path, self.common_dir_path, self.config_dir_path, self.user_config_path
    
    def get_wiliot_root_app_dir(self):
        return self.wiliot_root_path
    
    def get_common_dir(self):
        return self.common_dir_path
    
    def get_config_dir(self):
        return self.config_dir_path
    
    def get_user_config_file(self, client_type=None):
        if client_type is None:
            return self.user_config_path
        return self.user_config_path.replace('.json', f'_{client_type}.json')


def open_json(folder_path, file_path, default_values=None):
    """
    opens config json
    :type folder_path: string
    :param folder_path: the folder path which contains the desired file
    :type file_path: string
    :param file_path: the file path which contains the json
            (including the folder [file_path = folder_path+"json_file.json"])
    :type default_values: dictionary
    :param default_values: default values for the case of empty json
    :return: the desired json object
    """
    if not os.path.exists(folder_path):
        pathlib.Path(folder_path).mkdir(parents=True, exist_ok=True)
    
    file_exists = os.path.isfile(file_path)
    if not file_exists or os.stat(file_path).st_size == 0:
        # save the default values to json
        with open(file_path, "w") as out_file:
            json.dump(default_values, out_file)
        
        return json.load(open(file_path, "rb"))
    else:
        with open(file_path) as f:
            json_content = f.read()
        if len(json_content) == 0:
            with open(file_path, "w") as out_file:
                json.dump(default_values, out_file)
            json_content = json.load(open(file_path, "rb"))
        else:
            json_content = json.loads(json_content)
        return json_content


def valid_packet_start(msg):
    '''
    Function to check if the packet starts with 'process_packet' or 'full_packet', if it is, it will return True
    If
    '''
    pattern = r'(' + '|'.join(PACKET_PREFIXES_MAPPING) + r')\("(.*?)"\)'
    match = re.search(pattern, msg)

    return match.group(2) if match else ''


def check_user_config_is_ok(owner_id=None, env=None, client_type=None):
    raise NotImplementedError('This gui is deprecated, use GetApiKey instead')


def csv_to_dict(path=None):
    """
    convert csv to dictionary (arranged by columns)
    :param path: the csv path
    :type path: str
    :return: the data as dictionary
    :rtype: dict
    """
    if path:
        with open(path, 'r') as f:
            reader = csv.DictReader(f, delimiter=',')
            col_names = reader.fieldnames
            data_out = {col: [] for col in col_names}
            try:
                for row in reader:
                    for col in col_names:
                        data_out[col].append(row[col])
            except Exception as e:
                print("couldn't load csv due to {}".format(e))
        return data_out
    else:
        print('please provide a path')
        return None


def set_logger(app_name, dir_name='', file_name='', common_run_name=None, folder_name=''):
    """
    logger path would be dir_name/app_name/folder_name/filename_time.log
    """
    if not isinstance(app_name, str) or app_name == '':
        raise ValueError(f'app_name must be a string, better to use the actuall application name: {app_name}')
    logger = logging.getLogger(app_name)

    formatter = logging.Formatter('%(asctime)s,%(msecs)d %(name)s %(levelname)s: %(message)s', '%Y-%m-%d %H:%M:%S')
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.DEBUG)
    stream_handler.setFormatter(formatter)
    wiliot_dir = WiliotDir()
    if dir_name == '':
        dir_name = re.sub(r'(?<!^)(?=[A-Z])', '_', app_name).lower()

    logger_path = os.path.join(wiliot_dir.get_wiliot_root_app_dir(), dir_name)
    if folder_name:
        logger_path = os.path.join(logger_path, folder_name)

    if not os.path.isdir(logger_path):
        os.makedirs(logger_path)

    if common_run_name is None:
        if file_name == '':
            file_name = dir_name + '_log'
        common_run_name = f'{file_name}_{datetime.datetime.now().strftime("%Y%m%d_%H%M%S")}'
    else:
        logger_path = os.path.join(logger_path, common_run_name)
        if not os.path.isdir(logger_path):
            os.mkdir(logger_path)

    logger_path = os.path.join(logger_path, f'{common_run_name}.log')
    file_handler = logging.FileHandler(logger_path, mode='a')
    file_formatter = logging.Formatter('%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s', '%H:%M:%S')
    file_handler.setFormatter(file_formatter)
    logger.addHandler(file_handler)
    logger.addHandler(stream_handler)
    logger.setLevel(logging.INFO)
    return logger_path, logger


class GetApiKey(object):
    def __init__(self, owner_id=None, env=None, client_type=None, gui_type='ttk'):
        self.owner_id = owner_id
        self.env = env
        self.fix_env_string()
        self.api_key_values = {}
        self.user_config_file_path = self.get_config_path(client_type)
        auth_gui_is_needed, api_key, cfg_data = self.check_file()
        if auth_gui_is_needed:
            gui_type = gui_type if gui_type is not None else 'ttk'
            if gui_type == 'ttk':
                self.popup_key(client_type, cfg_data)
            else:
                raise NotImplementedError(f'gui type: {gui_type} is not supported please select: ttk')
        else:
            self.api_key_values = {'api_key': api_key, 'owner_id': self.owner_id, 'env': self.env}

    def get_api_key(self):
        return self.api_key_values['api_key'] if 'api_key' in self.api_key_values else ''

    def get_config_path(self, client_type):
        env_dirs = WiliotDir()
        config_dir_path = env_dirs.get_config_dir()
        if not os.path.isdir(config_dir_path):
            pathlib.Path(config_dir_path).mkdir(parents=True, exist_ok=True)

        return env_dirs.get_user_config_file(client_type)

    def fix_env_string(self):
        if self.env is not None:
            self.env = self.env.lower()
        if self.env == '':
            self.env = 'prod'
        if self.env == 'non-prod':
            self.env = 'test'

    def check_file(self):
        auth_gui_is_needed = False
        api_key = None
        cfg_data = []

        if os.path.exists(self.user_config_file_path):
            cfg_data = open_json(folder_path=os.path.dirname(self.user_config_file_path),
                                 file_path=self.user_config_file_path)
            try:
                api_key_list = cfg_data['api_key']
                env_list = cfg_data['env']
                owner_list = cfg_data['owner_id']
                env_list = [env_element if env_element != 'non-prod' else 'test' for env_element in env_list]
                if not isinstance(api_key_list, list) or api_key_list == []:
                    print('api key is missing. Please enter it manually')
                    auth_gui_is_needed = True
                elif (self.owner_id is not None and self.owner_id not in owner_list) or \
                        (self.env is not None and self.env not in env_list):
                    print('api key does not match the request owner id or env. Please enter it manually')
                    auth_gui_is_needed = True
                else:
                    api_key = [k for i, k in enumerate(api_key_list)
                               if (self.owner_id is None or self.owner_id == owner_list[i]) and
                               (self.env is None or self.env == env_list[i])]
                    if api_key == []:
                        print('could not find the api key for the specified environment and owner id')
                        auth_gui_is_needed = True
                    api_key = api_key[0]
            except Exception as e:
                auth_gui_is_needed = True
                print("Config file is not readable at path {}. Exception {}\n Please enter new credentials".format(
                    self.user_config_file_path, e))
        else:
            print("Config file user_configs.json doesn't exist {}\n".format(self.user_config_file_path))
            auth_gui_is_needed = True

        return auth_gui_is_needed, api_key, cfg_data

    def popup_key(self, client_type, cfg_data):
        default_font = ("Helvetica", 10)
        popup = Tk()
        popup.eval('tk::PlaceWindow . center')
        popup.wm_title('Missing API key')
        client_type = client_type if client_type is not None else 'DEFAULT'

        def quit_tester():
            try:
                popup.destroy()
            except Exception as e:
                print(e)
                exit(1)

        popup.protocol("WM_DELETE_WINDOW", quit_tester)

        def ok():
            api_key = c1.get()
            self.api_key_values = {'api_key': api_key, 'owner_id': self.owner_id, 'env': self.env}
            self.save_api_key(cfg_data)
            popup.destroy()

        l1 = Label(popup, text=f'Get API KEY for:\n\n'
                               f'ClientType: {client_type}\nOwner: {self.owner_id}\nEnv: {self.env}', font=default_font)
        l1.grid(row=2, column=0, padx=10, pady=10, columnspan=5)

        l4 = Label(popup, text='API Key:', font=default_font)
        l4.grid(row=6, column=0, padx=10, pady=10)
        c1 = ttk.Combobox(popup, state='normal', width=30)
        c1.grid(row=7, column=0, padx=10, pady=15)
        b3 = Button(popup, text="OK", command=ok, height=1, width=10)
        b3.grid(row=8, column=0, padx=10, pady=10)
        popup.mainloop()

    def save_api_key(self, cfg_data):
        api_key = self.get_api_key()
        if api_key:
            with open(self.user_config_file_path, 'w') as cfg:
                if isinstance(cfg_data, dict):
                    for k in self.api_key_values.keys():
                        if k in cfg_data.keys() and isinstance(cfg_data[k], list):
                            cfg_data[k].append(self.api_key_values[k])
                        else:
                            cfg_data[k] = [self.api_key_values[k]]
                    json.dump(cfg_data, cfg)
                else:
                    value_to_save = {k: [v] for k, v in self.api_key_values.items()}
                    json.dump(value_to_save, cfg)
        else:
            print('api key is missing. Please try again\n')


def open_directory(path):
    if sys.platform == 'darwin':
        subprocess.run(["open", path])
    elif 'win' in sys.platform:
        os.startfile(path)
    else:
        subprocess.run(["xdg-open", path])


if __name__ == '__main__':
    k = GetApiKey(owner_id='hi', env='test')
    print(k.get_api_key())



"""
Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

from wiliot_core.packet_data.config_files.data_mapping.packet_data_info import *
from wiliot_core.packet_data.config_files.data_mapping.gw_data_info import *


def extract_packet_data_types():
    return {key: value['type'] for key, value in packet_data_info.items()}


def extract_gw_data_types():
    return {key: value['type'] for key, value in gw_data_info.items()}


def check_packet_length_with_no_indication(packet_data_input):
    expected_length = None
    received_len = len(packet_data_input)
    for packet_prefix, length_dict in packet_length_types.items():
        if received_len == length_dict['packet_tag_length']:
            packet_data_input += '0' * (
                    find_len(gw_data_info, 'stat_param') + find_len(gw_data_info, 'rssi'))
            expected_length = length_dict

        elif received_len == length_dict['packet_tag_length'] + (
                find_len(gw_data_info, 'stat_param') + find_len(gw_data_info, 'rssi')):
            expected_length = length_dict

        elif length_dict['name'] == 'LEGACY':
            if received_len == length_dict['packet_tag_length'] - find_len(packet_data_info, 'header'):
                packet_data_input = packet_prefix + packet_data_input + '0' * (
                        find_len(gw_data_info, 'stat_param') + find_len(gw_data_info,
                                                                                                    'rssi'))
                expected_length = length_dict
            elif received_len == \
                    length_dict['packet_tag_length'] - find_len(packet_data_info, 'header') + (
                    find_len(gw_data_info, 'stat_param') + find_len(gw_data_info, 'rssi')):
                packet_data_input = packet_prefix + packet_data_input
                expected_length = length_dict

        if expected_length is not None:
            break

    return packet_data_input, expected_length


def fix_packet_length(packet_data_input, is_full_packet):
    for full_key in packet_length_types.keys():
        if packet_data_input.startswith(full_key):
            is_full_packet = True

    if is_full_packet is None:
        packet_data_input, expected_length = check_packet_length_with_no_indication(packet_data_input)
        if expected_length is None:
            raise Exception(f'invalid packet length for packet {packet_data_input}, '
                            f'these are the valid tag packet length: {packet_length_types}')
    elif is_full_packet:
        expected_length = packet_length_types[packet_data_input[:find_len(packet_data_info, 'header')]]
    else:
        packet_data_input = '4225' + packet_data_input
        expected_length = packet_length_types[packet_data_input[:find_len(packet_data_info, 'header')]]
    
    # check if packet length is valid
    received_len = len(packet_data_input)

    if received_len == expected_length['packet_tag_length'] + GW_PACKET_LEN:
        pass
    elif received_len == expected_length['packet_tag_length']:
        packet_data_input += '0' * GW_PACKET_LEN
    else:
        raise Exception(f'invalid packet length for packet {packet_data_input}, '
                        f'expected tag packet length: {expected_length["packet_tag_length"]}')
    
    return packet_data_input, expected_length


def parser(data_dict, packet_data_in, general_args, array_data=False):
    result = {'packet_data': packet_data_in}
    packet_type = general_args['expected_length']['name']
    for key, info in data_dict.items():
        if key in result.keys():
            continue
            
        input_data = result.get(info['input'])
        start, length = info['index_info'].get(packet_type, (0, 0)) if packet_type in info['index_info'] else info['index_info'].get('all', (0, 0))
        func = info.get('function')
        args_names = info.get('args', [])
        data_type = info.get('type')
        data_out = None
        if input_data is not None:
            data_out = input_data[start:start + length]

        if callable(func):
            args_g = {arg_name: general_args[arg_name] for arg_name in args_names if arg_name in general_args}
            args_r = {arg_name: result[arg_name] for arg_name in args_names if arg_name in result}
            result[key] = func(data_out, args={**args_g, **args_r})
        else:
            result[key] = data_out
        
        if type(result[key]) != data_type:
            raise Exception(f'key: {key} has wrong type {type(result[key])} (should be {data_type})')
    
    del result['packet_data']
    if array_data:
        for k, v in result.items():
            result[k] = np.array([v])
    return result


def parse_packet(packet_data, time_from_start, is_full_packet, ignore_crc):
    # update packet string and check its type
    packet_data_fixed, expected_length = fix_packet_length(packet_data, is_full_packet)  # This should be the only function here
    # arrange packet_params
    packet_params = {
        'expected_length': expected_length, 
        'ignore_crc': ignore_crc,
        'time_from_start': time_from_start,
        }

    packet_result = parser(packet_data_info, packet_data_fixed, packet_params)
    gw_result = parser(gw_data_info, packet_data_fixed, packet_params, array_data=True)

    return gw_result, packet_result

def construct_packet_from_fields(packet_data, fields_to_zero=None, fields_to_convert=None, gw_data='', prefix=False,
                                 gw_fields_to_convert=None):
    """
    fields_to_zero : a list of fields names, need to be zero
    fields_to_convert : dict of field name as key and its value based on packet_structure variable
    gw_fields_to_convert : if gw_data is not specified,
                           dict of field name as key and its value based on gw_bit_structure variable
    """
    new_packet = []
    is_legacy_packet = packet_length_types.get(packet_data['header'], {'name': 'LEGACY'})['name'] == 'LEGACY'
    for field in packet_structure:
        if is_legacy_packet and field in BLE5_ONLY_FIELDS:
            continue  # do not add ble5 fields to legacy packets
        if fields_to_zero and field in fields_to_zero:
            field_len = find_len(packet_data_info, field)
            new_packet.append('0' * field_len)  # BLE5-EXT and BLE5-DBL-EXT have the same length
        elif fields_to_convert and field in fields_to_convert:
            new_packet.append(fields_to_convert[field].upper())
        else:
            new_packet.append(packet_data.get(field, ''))

    construct_packet = ''.join(new_packet)
    if gw_data == '' and gw_fields_to_convert is not None:
        gw_bit_data = ''
        if 'crc_valid' not in gw_fields_to_convert.keys():
            gw_fields_to_convert['crc_valid'] = 1
        for field, field_bit_len in gw_bit_structure.items():
            if field in gw_fields_to_convert.keys():
                bin_field = bin(int(gw_fields_to_convert[field]))[2:]
            else:
                bin_field = '0'
            bin_field = bin_field.zfill(field_bit_len)
            bin_field = bin_field[-field_bit_len:]
            gw_bit_data += str(bin_field)
        gw_data = hex(int(gw_bit_data, 2))[2:].upper()

    construct_packet += gw_data

    if prefix:
        construct_packet = f'process_packet("{construct_packet}")' if is_legacy_packet \
            else f'full_packet("{construct_packet}")'
    return construct_packet



"""
Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""
from wiliot_core.local_gateway.local_gateway_core import valid_output_power_vals, valid_bb, \
    valid_sub1g_output_power


class CustomEnergyPattern(object):
    def __init__(self, gw_obj):
        self.gw_obj = gw_obj
        self.power_2_4 = valid_output_power_vals
        self.abs_power_arr = [p['abs_power'] for p in self.power_2_4]

    def set_scan_channel(self, scan_ch):
        """
        set scanning channel
        :param scan_ch: can be ble ch (0-40) or frequency in Mhz
        :type scan_ch: int
        :return: the gw ack response
        :rtype: dict with time and raw fields
        """
        rsp = self.gw_obj.write('!scan_ch {} 37'.format(scan_ch))
        return rsp

    def set_time_period_2_4_ghz(self, period_2_4):
        """
        setting the time period of the gateway application (tx-rx)
        :param period_2_4: the time period for the 2.4ghz tx/rx (between 0-255) in milliseconds
        :type period_2_4: int
        :return: the gw ack response
        :rtype: dict with time and raw fields
        """
        rsp = self.gw_obj.write('!set_2_4_ghz_time_period {}'.format(period_2_4))
        return rsp

    def set_beacons_pattern_2_4_ghz(self, beacon_2_4_frequencies, beacon_2_4_duration,
                                    beacon_to_beacon=None, beacon_to_energy=None):
        """
        set the beacon pattern in 2.4GHz including number of beacons, frequencies and timing
        :param beacon_2_4_frequencies: a list of frequencies, the number of beacons equals to the length of this list
        :type beacon_2_4_frequencies: list
        :param beacon_2_4_duration: the beacon duration in microseconds
        :type beacon_2_4_duration: int
        :param beacon_to_beacon: the time between two successive beacons in microseconds
        :type beacon_to_beacon: int
        :param beacon_to_energy: relevant only if beacon_to_beacon was specified.
                                 the time between the last beacon and the energy in microseconds
        :type beacon_to_energy: int
        :return: the gw ack response
        :rtype: dict with time and raw fields
        """
        cmd = '!set_beacons_pattern {} {}'.format(beacon_2_4_duration, len(beacon_2_4_frequencies))
        for f in beacon_2_4_frequencies:
            cmd += ' {}'.format(f)
        if beacon_to_beacon is not None:
            cmd += ' {}'.format(beacon_to_beacon)
            if beacon_to_energy is not None:
                cmd += ' {}'.format(beacon_to_energy)
        rsp = self.gw_obj.write(cmd)
        return rsp

    def set_beacons_output_power(self, beacon_2_4_power):
        """
        set the beacons output power
        :param beacon_2_4_power: the beacon output power according to the valid list of 2.4ghz output power
        :type beacon_2_4_power:
        :return: the gw ack response
        :rtype: dict with time and raw fields
        """
        beacon_backoff = self.power_2_4[-1]['abs_power'] - beacon_2_4_power
        if beacon_backoff not in valid_bb:
            print('{} power value is not a valid power value (valid power values are: {})'.format(beacon_2_4_power,
                                                                                                  self.abs_power_arr))
        rsp = self.gw_obj.write('!beacons_backoff {}'.format(self.power_2_4[-1]['abs_power'] - beacon_2_4_power))
        return rsp

    def set_energy_pattern_2_4_ghz(self, energy_2_4_frequencies, is_sub1g_energy, energy_2_4_duration=None):
        """
        set the energy pattern in 2.4ghz, including frequencies and timing
        :param energy_2_4_frequencies: list of frequency,
                                       so each cycle the next frequency will be used for the energizing frequency
        :type energy_2_4_frequencies: list
        :param is_sub1g_energy: true if energy in sub1g is also needed
        :type is_sub1g_energy: bool
        :param energy_2_4_duration: the time duration of the energy constant wave in microsecond
        :type energy_2_4_duration: int
        :return: the gw ack response
        :rtype: dict with time and raw fields
        """
        cmd = '!set_dyn_energizing_pattern 6 {} {}'.format(is_sub1g_energy, len(energy_2_4_frequencies))
        for f in energy_2_4_frequencies:
            cmd += ' {}'.format(f)
        if energy_2_4_duration is not None:
            cmd += ' {}'.format(energy_2_4_duration)
        rsp = self.gw_obj.write(cmd)
        return rsp

    def set_energy_output_power_2_4_ghz(self, energy_2_4_power):
        """
        set the output power of the energy
        :param energy_2_4_power: the constant wave output power according to the valid list of 2.4ghz output power
        :type energy_2_4_power: int
        :return: the gw ack response
        :rtype: list of two dictionaries, each dict with time and raw fields
        """
        if energy_2_4_power not in self.abs_power_arr:
            print('{} power value is not a valid power value (valid power values are: {})'.format(energy_2_4_power,
                                                                                                  self.abs_power_arr))
        abs_output_power_index = self.abs_power_arr.index(energy_2_4_power)
        rsp1 = self.gw_obj.write('!bypass_pa {}'.format(self.power_2_4[abs_output_power_index]['bypass_pa']))
        rsp2 = self.gw_obj.write('!output_power {}'.format(self.power_2_4[abs_output_power_index]['gw_output_power']))
        return [rsp1, rsp2]

    def set_time_period_sub1g(self, period_sub1g):
        """
        set the time period of the sub1g tx/rx
        :param period_sub1g: the time period of the sub1g tx/rx in milliseconds
        :type period_sub1g: int
        :return: the gw ack response
        :rtype: dict with time and raw fields
        """
        rsp = self.gw_obj.write('!set_sub_1_ghz_time_period {}'.format(period_sub1g))
        return rsp

    def set_energy_pattern_sub1g(self, energy_sub1g_frequencies, energy_sub1g_duration=None):
        """
        set the energy pattern in sub1g, including frequencies and timing
        :param energy_sub1g_frequencies: list of frequency,
                                         each cycle the next frequency will be used for the energizing frequency
        :type energy_sub1g_frequencies: list
        :param energy_sub1g_duration: the time duration of the constant wave in microseconds
        :type energy_sub1g_duration: int
        :return: the gw ack response
        :rtype: dict with time and raw fields
        """
        cmd = '!set_sub_1_ghz_energy_params {}'.format(len(energy_sub1g_frequencies))
        for f in energy_sub1g_frequencies:
            cmd += ' {}'.format(f)
        if energy_sub1g_duration is not None:
            cmd += '{}'.format(energy_sub1g_duration)
        rsp = self.gw_obj.write(cmd)
        return rsp

    def set_energy_output_power_sub1g(self, energy_sub1g_power):
        """
        set the energy output power in sub1g
        :param energy_sub1g_power: should be between 17-29 dbm
        :type energy_sub1g_power: int
        :return: the gw ack response
        :rtype: dict with time and raw fields
        """
        if energy_sub1g_power not in valid_sub1g_output_power:
            print('{} power value is not a valid power value (valid power values are: {})'.format(
                energy_sub1g_power, valid_sub1g_output_power))
        rsp = self.gw_obj.write('!set_sub_1_ghz_power {}'.format(energy_sub1g_power))
        return rsp

    def run_gateway_app(self):
        """
        start receiving and transmitting
        :return: the gw ack response
        :rtype: dict with time and raw fields
        """
        rsp = self.gw_obj.write('!gateway_app')
        return rsp

    def stop_gateway_app(self):
        """
        stop receiving and transmitting
        :return: the gw ack response
        :rtype: dict with time and raw fields
        """
        rsp = self.gw_obj.write('!cancel')
        return rsp


if __name__ == '__main__':
    from wiliot_core.local_gateway.local_gateway_core import WiliotGateway, ActionType
    import time

    gw = WiliotGateway(port='COM3', auto_connect=True)  # generate the gw obj and connect to it
    custom_ep = CustomEnergyPattern(gw_obj=gw)  # init custom ep
    if gw.connected:
        gw.start_continuous_listener()  # start a thread for getting the packets from the gw
        rsp = custom_ep.set_scan_channel(scan_ch=37)  # config the gw using the custom ep functions
        print(rsp)
        rsp = custom_ep.run_gateway_app()  # run the gw app
        print(rsp)
        print('energizing the tag for 30 seconds')
        time.sleep(30)
        packets = gw.get_packets(action_type=ActionType.ALL_SAMPLE)
        for p in packets:  # printing packets
            print(p.get_packet_string())
        rsp = custom_ep.stop_gateway_app()  # stop the gw app
        print(rsp)
    else:
        print('gw is not connected, please check connection and rerun')
    gw.exit_gw_api()  # close all communication with gw ap



from enum import Enum


class OutputPowers(Enum):
    pos0dbm = {'output_power': 'RADIO_TXPOWER_TXPOWER_0dBm', 'numeric_val': 0}
    pos2dbm = {'output_power': 'RADIO_TXPOWER_TXPOWER_Pos2dBm', 'numeric_val': 2}
    pos3dbm = {'output_power': 'RADIO_TXPOWER_TXPOWER_Pos3dBm', 'numeric_val': 3}
    pos4dbm = {'output_power': 'RADIO_TXPOWER_TXPOWER_Pos4dBm', 'numeric_val': 4}
    pos5dbm = {'output_power': 'RADIO_TXPOWER_TXPOWER_Pos5dBm', 'numeric_val': 5}
    pos6dbm = {'output_power': 'RADIO_TXPOWER_TXPOWER_Pos6dBm', 'numeric_val': 6}
    pos7dbm = {'output_power': 'RADIO_TXPOWER_TXPOWER_Pos7dBm', 'numeric_val': 7}
    pos8dbm = {'output_power': 'RADIO_TXPOWER_TXPOWER_Pos8dBm', 'numeric_val': 8}
    neg4dbm = {'output_power': 'RADIO_TXPOWER_TXPOWER_Neg4dBm', 'numeric_val': -4}
    neg8dbm = {'output_power': 'RADIO_TXPOWER_TXPOWER_Neg8dBm', 'numeric_val': -8}
    neg12dbm = {'output_power': 'RADIO_TXPOWER_TXPOWER_Neg12dBm', 'numeric_val': -12}
    neg16dbm = {'output_power': 'RADIO_TXPOWER_TXPOWER_Neg16dBm', 'numeric_val': -16}
    neg20dbm = {'output_power': 'RADIO_TXPOWER_TXPOWER_Neg20dBm', 'numeric_val': -20}
    neg30dbm = {'output_power': 'RADIO_TXPOWER_TXPOWER_Neg30dBm', 'numeric_val': -30}
    neg40dbm = {'output_power': 'RADIO_TXPOWER_TXPOWER_Neg40dBm', 'numeric_val': -40}
